"""Generate C# code for reporting errors by including the code directly."""

import io
import textwrap

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(namespace: csharp_common.NamespaceIdentifier) -> str:
    """
    Generate the C# code for reporting errors.

    The ``namespace`` defines the AAS C# namespace.
    """
    blocks = [
        Stripped(
            textwrap.dedent(
                f"""\
                /// <summary>
                /// Capture a path segment of a value in a model.
                /// </summary
                public abstract class Segment {{
                {I}// Intentionally empty.
                }}"""
            )
        ),
        Stripped(
            textwrap.dedent(
                f"""\
                public class NameSegment : Segment {{
                {I}internal readonly string Name;
                {I}internal NameSegment(string name)
                {I}{{
                {II}Name = name;
                {I}}}
                }}"""
            )
        ),
        Stripped(
            textwrap.dedent(
                f"""\
                public class IndexSegment : Segment {{
                {I}internal readonly int Index;
                {I}internal IndexSegment(int index)
                {I}{{
                {II}Index = index;
                {I}}}
                }}"""
            )
        ),
        Stripped(
            textwrap.dedent(
                f"""\
                internal static System.Text.RegularExpressions.Regex VariableNameRe = (
                {I}new  System.Text.RegularExpressions.Regex(
                {II}@"^[a-zA-Z_][a-zA-Z_0-9]*$"));"""
            )
        ),
        # We have to indent a lot so we do not use textwrap.dedent for better
        # readability.
        Stripped(
            f"""\
/// <summary>
/// Generate a JSON Path based on the path segments.
/// </summary>
/// <remarks>
/// See, for example, this page for more information on JSON path:
/// https://support.smartbear.com/alertsite/docs/monitors/api/endpoint/jsonpath.html
/// </remarks>
public static string GenerateJsonPath(
{I}ICollection<Segment> segments)
{{
{I}var parts = new List<string>(segments.Count);
{I}int i = 0;
{I}foreach(var segment in segments)
{I}{{
{II}string? part = null;
{II}switch (segment)
{II}{{
{III}case NameSegment nameSegment:
{IIII}if (VariableNameRe.IsMatch(nameSegment.Name))
{IIII}{{
{IIIII}part = (i == 0) ? nameSegment.Name : $".{{nameSegment.Name}}";
{IIII}}}
{IIII}else
{IIII}{{
{IIIII}string escaped = nameSegment.Name
{IIIIII}.Replace("\\\\", "\\\\\\\\")
{IIIIII}.Replace("\\"", "\\\\\\"")
{IIIIII}.Replace("\\b", "\\\\b")
{IIIIII}.Replace("\\f", "\\\\f")
{IIIIII}.Replace("\\n", "\\\\n")
{IIIIII}.Replace("\\r", "\\\\r")
{IIIIII}.Replace("\\t", "\\\\t");
{IIIII}part = $"[\\"{{escaped}}\\"]";
{IIII}}}
{IIII}break;
{III}case IndexSegment indexSegment:
{IIII}part = $"[{{indexSegment.Index}}]";
{IIII}break;
{III}default:
{IIII}throw new System.InvalidOperationException(
{IIIII}$"Unexpected segment type: {{segment.GetType()}}");
{II}}}
{II}parts.Add(part);
{I}}}
{I}return string.Join("", parts);
}}"""
        ),
        Stripped(
            textwrap.dedent(
                f"""\
                /// <summary>
                /// Represent an error during the deserialization or the verification.
                /// </summary>
                public class Error
                {{
                {I}internal LinkedList<Segment> _pathSegments = new LinkedList<Segment>();
                {I}public readonly string Cause;
                {I}public ICollection<Segment> PathSegments {{
                {II}get {{ return _pathSegments; }}
                {I}}}
                {I}internal Error(string cause)
                {I}{{
                {II}Cause = cause;
                {I}}}
                }}"""
            )
        ),
    ]

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            f"""\
            namespace {namespace}
            {{
            {I}/// <summary>
            {I}/// Provide reporting for de/serialization and verification.
            {I}/// </summary>
            {I}public static class Reporting
            {I}{{
            """
        )
    )

    for i, deserialize_block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(deserialize_block, II))

    writer.write(f"\n{I}}}  // public static class Reporting")
    writer.write(f"\n}}  // namespace {namespace}")

    # pylint: disable=line-too-long
    blocks = [
        csharp_common.WARNING,
        Stripped("using System.Collections.Generic;  // can't alias"),
        Stripped(f"using Aas = {namespace};"),
        Stripped(writer.getvalue()),
        csharp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Infer constraints representable in common schemas such as JSON Schema or XSD."""

from aas_core_codegen.infer_for_schema import _len, _pattern, _inline, _stringify

LenConstraint = _len.LenConstraint

PatternConstraint = _pattern.PatternConstraint

ConstraintsByProperty = _inline.ConstraintsByProperty
infer_constraints_by_class = _inline.infer_constraints_by_class

dump = _stringify.dump
dump_len_constraints_by_properties = _stringify.dump_len_constraints_by_properties
dump_patterns = _stringify.dump_patterns
dump_patterns_by_properties = _stringify.dump_patterns_by_properties



"""
Transpile the meta-model into C# and compile them into a project.

This live tests expects dotnet to be installed on the machine.
"""
import io
import os
import pathlib
import subprocess
import sys
import tempfile
import textwrap

import aas_core_codegen.main


def main() -> int:
    """Execute the main routine."""
    print("Running dotnet --version to check that dotnet is available...")
    exit_code = subprocess.call(
        ["dotnet", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )
    if exit_code != 0:
        print(
            f"Failed to execute ``dotnet --version`` "
            f"with the exit code {exit_code}. Is dotnet installed?",
            file=sys.stderr,
        )
        return 1

    with tempfile.TemporaryDirectory() as tmp_dir:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

        parent_case_dir = repo_dir / "test_data" / "csharp" / "test_main"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        for case_dir in parent_case_dir.iterdir():
            assert case_dir.is_dir(), case_dir

            model_pth = case_dir / "input/meta_model.py"
            assert model_pth.exists() and model_pth.is_file(), model_pth

            snippets_dir = case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            output_dir = pathlib.Path(tmp_dir) / case_dir.name

            print(
                f"Generating the files based on the case {case_dir} "
                f"to: {output_dir} ..."
            )

            output_dir.mkdir(exist_ok=True, parents=True)

            params = aas_core_codegen.main.Parameters(
                model_path=model_pth,
                target=aas_core_codegen.main.Target.CSHARP,
                snippets_dir=snippets_dir,
                output_dir=output_dir,
            )

            stdout = io.StringIO()
            stderr = io.StringIO()

            return_code = aas_core_codegen.main.execute(
                params=params, stdout=stdout, stderr=stderr
            )

            assert (
                stderr.getvalue() == ""
            ), f"Expected no stderr on valid models, but got:\n{stderr.getvalue()}"

            assert (
                return_code == 0
            ), f"Expected return code 0 on valid models, but got: {return_code}"

            print("Generating the .csproj file...")
            csproj_pth = output_dir / "SomeProject.csproj"
            csproj_pth.write_text(
                textwrap.dedent(
                    """\
                <Project Sdk="Microsoft.NET.Sdk">
                    <PropertyGroup>
                        <TargetFramework>net6.0</TargetFramework>
                        <Nullable>enable</Nullable>
                        <Configurations>Debug;Release;DebugSlow</Configurations>
                        <Platforms>AnyCPU</Platforms>
                        <LangVersion>8</LangVersion>
                    </PropertyGroup>
                </Project>
                """
                ),
                encoding="utf-8",
            )

            print("Calling dotnet build...")
            exit_code = subprocess.call(["dotnet", "build", "."], cwd=str(output_dir))
            if exit_code != 0:
                print(
                    f"ERROR: Expected the build to succeed, "
                    f"but got exit code: {exit_code}",
                    file=sys.stderr,
                )
                return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



# pylint: disable=missing-docstring

import os
import pathlib
import textwrap
import unittest
from typing import List, Tuple

import tests.common
from aas_core_codegen import intermediate
from aas_core_codegen.intermediate import doc as intermediate_doc
from aas_core_codegen.common import Identifier


class Test_in_lining_of_constructor_statements(unittest.TestCase):
    def test_case(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class VeryAbstract:
                some_property: int

                @require(lambda some_property: some_property > 0)
                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property


            @abstract
            class Abstract(VeryAbstract):
                another_property: int

                @require(lambda another_property: another_property > 0)
                def __init__(self, some_property: int, another_property: int) -> None:
                    VeryAbstract.__init__(self, some_property)
                    self.another_property = another_property

            class Concrete(Abstract):
                yet_another_property: int

                @require(lambda yet_another_property: yet_another_property > 0)
                def __init__(
                        self,
                        some_property: int,
                        another_property: int,
                        yet_another_property: int
                ) -> None:
                    Abstract.__init__(self, some_property, another_property)
                    self.yet_another_property = yet_another_property


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        concrete = symbol_table.must_find(Identifier("Concrete"))
        assert isinstance(concrete, intermediate.Class)

        self.assertEqual(
            ["some_property", "another_property", "yet_another_property"],
            [stmt.name for stmt in concrete.constructor.statements],
        )


class Test_parsing_docstrings(unittest.TestCase):
    def test_class_reference(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                This is some documentation.

                Nested reference :class:`.Some_class`
                """

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        some_class = symbol_table.must_find(Identifier("Some_class"))
        assert isinstance(some_class, intermediate.Class)

        assert some_class.description is not None
        assert len(some_class.description.remarks) == 1

        symbol_references = list(
            some_class.description.remarks[0].findall(
                condition=intermediate_doc.SymbolReference
            )
        )

        self.assertEqual(1, len(symbol_references))
        self.assertIsInstance(symbol_references[0].symbol, intermediate.Class)

    def test_constraint_and_constraintref(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                This is some documentation.

                See :constraintref:`AAS-001`.

                :constraint AAS-001:
                    some constraint
                """

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            '''
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        some_class = symbol_table.must_find(Identifier("Some_class"))
        assert isinstance(some_class, intermediate.Class)

        assert some_class.description is not None
        assert len(some_class.description.remarks) == 1

        constraint_references = list(
            some_class.description.remarks[0].findall(
                condition=intermediate_doc.ConstraintReference
            )
        )
        self.assertEqual(1, len(constraint_references))
        self.assertEqual("AAS-001", constraint_references[0].reference)

        self.assertListEqual(
            ["AAS-001"], list(some_class.description.constraints_by_identifier.keys())
        )


class Test_against_recorded(unittest.TestCase):
    RERECORD = os.environ.get("AAS_CORE_CODEGEN_RERECORD", "").lower() in (
        "1",
        "true",
        "on",
    )

    def test_cases(self) -> None:
        this_dir = pathlib.Path(os.path.realpath(__file__)).parent
        repo_root = this_dir.parent.parent
        test_cases_dir = repo_root / "test_data/intermediate"

        assert test_cases_dir.exists(), f"{test_cases_dir=}"
        assert test_cases_dir.is_dir(), f"{test_cases_dir=}"

        # The expected cases should have no errors.
        expected_pths = sorted((test_cases_dir / "expected").glob("**/meta_model.py"))
        meta_model_pths_expected_errors = [
            (pth, False) for pth in expected_pths
        ]  # type: List[Tuple[pathlib.Path, bool]]

        unexpected_pths = sorted(
            (test_cases_dir / "unexpected").glob("**/meta_model.py")
        )

        meta_model_pths_expected_errors.extend((pth, True) for pth in unexpected_pths)

        for meta_model_pth, expected_errors in meta_model_pths_expected_errors:
            case_dir = meta_model_pth.parent

            try:
                source = meta_model_pth.read_text(encoding="utf-8")
            except Exception as exception:
                raise AssertionError(
                    f"Unexpected exception when reading "
                    f"from {meta_model_pth.relative_to(repo_root)}"
                ) from exception

            try:
                symbol_table, error = tests.common.translate_source_to_intermediate(
                    source=source
                )
            except Exception as exception:
                raise AssertionError(
                    f"Unexpected exception in source-to-intermediate translation "
                    f"for source {meta_model_pth.relative_to(repo_root)}"
                ) from exception

            if not expected_errors and error is not None:
                raise AssertionError(
                    f"Expected no errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got:\n"
                    f"{tests.common.most_underlying_messages(error)}"
                )

            elif expected_errors and error is None:
                raise AssertionError(
                    f"Expected errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got none."
                )

            else:
                pass

            expected_symbol_table_pth = case_dir / "expected_symbol_table.txt"
            expected_error_pth = case_dir / "expected_error.txt"

            if expected_errors:
                if expected_symbol_table_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded symbol table file when errors "
                        f"are expected: {expected_symbol_table_pth}"
                    )

                assert error is not None

                error_str = tests.common.most_underlying_messages(error)

                if Test_against_recorded.RERECORD:
                    expected_error_pth.write_text(error_str, encoding="utf-8")
                else:
                    expected_error_str = expected_error_pth.read_text(encoding="utf-8")
                    self.assertEqual(expected_error_str, error_str, f"{case_dir=}")

            else:
                if expected_error_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded error file when no errors "
                        f"are expected: {expected_error_pth}"
                    )

                assert symbol_table is not None

                symbol_table_str = intermediate.dump(symbol_table)

                if Test_against_recorded.RERECORD:
                    expected_symbol_table_pth.write_text(
                        symbol_table_str, encoding="utf-8"
                    )
                else:
                    try:
                        expected_symbol_table_str = expected_symbol_table_pth.read_text(
                            encoding="utf-8"
                        )
                    except Exception as exception:
                        raise RuntimeError(
                            f"Failed to read the file representing "
                            f"the expected symbol table: {expected_symbol_table_pth}"
                        ) from exception

                    self.assertEqual(
                        expected_symbol_table_str,
                        symbol_table_str,
                        f"{case_dir=}, {error=}",
                    )


if __name__ == "__main__":
    unittest.main()



"""Provide types of the intermediate representation."""
import abc
import enum
import pathlib
from typing import (
    Sequence,
    Optional,
    Union,
    TypeVar,
    Mapping,
    MutableMapping,
    Final,
    FrozenSet,
    Set,
    Tuple,
    OrderedDict,
)

import docutils.nodes
from icontract import require, invariant, ensure, DBC

from aas_core_codegen import parse
from aas_core_codegen.common import (
    Identifier,
    assert_never,
    assert_union_of_descendants_exhaustive,
    assert_union_without_excluded,
)
from aas_core_codegen.intermediate import construction
from aas_core_codegen.parse import tree as parse_tree

_MODULE_NAME = pathlib.Path(__file__).parent.name


class PrimitiveType(enum.Enum):
    """List primitive types."""

    BOOL = "bool"
    INT = "int"
    FLOAT = "float"
    STR = "str"
    BYTEARRAY = "bytearray"


assert sorted(literal.value for literal in PrimitiveType) == sorted(
    parse.PRIMITIVE_TYPES
), "All primitive types specified in the intermediate layer"

STR_TO_PRIMITIVE_TYPE = {
    literal.value: literal for literal in PrimitiveType
}  # type: Mapping[str, PrimitiveType]


class TypeAnnotation(DBC):
    """Represent a general type annotation."""

    #: Relation to the parse stage
    parsed: Final[parse.TypeAnnotation]

    def __init__(self, parsed: parse.TypeAnnotation) -> None:
        """Initialize with the given values."""
        self.parsed = parsed

    @abc.abstractmethod
    def __str__(self) -> str:
        # Signal that this class is a purely abstract one
        raise NotImplementedError()


class PrimitiveTypeAnnotation(TypeAnnotation):
    """Represent a primitive type such as ``int``."""

    def __init__(self, a_type: PrimitiveType, parsed: parse.TypeAnnotation) -> None:
        """Initialize with the given values."""
        TypeAnnotation.__init__(self, parsed=parsed)
        self.a_type = a_type

    def __str__(self) -> str:
        return str(self.a_type.value)


class OurTypeAnnotation(TypeAnnotation):
    """
    Represent an atomic annotation defined by a symbol in the meta-model.

     For example, ``Asset``.
    """

    def __init__(self, symbol: "Symbol", parsed: parse.TypeAnnotation) -> None:
        """Initialize with the given values."""
        TypeAnnotation.__init__(self, parsed=parsed)
        self.symbol = symbol

    def __str__(self) -> str:
        return self.symbol.name


class ListTypeAnnotation(TypeAnnotation):
    """Represent a type annotation involving a ``List[...]``."""

    def __init__(self, items: "TypeAnnotationUnion", parsed: parse.TypeAnnotation):
        TypeAnnotation.__init__(self, parsed=parsed)

        self.items = items

    def __str__(self) -> str:
        return f"List[{self.items}]"


# NOTE (mristin, 2021-11-19):
# We do not support other generic types except for ``List``. In the future we might
# add support for ``Set``, ``MutableMapping`` *etc.*


class OptionalTypeAnnotation(TypeAnnotation):
    """Represent a type annotation involving an ``Optional[...]``."""

    def __init__(self, value: "TypeAnnotationUnion", parsed: parse.TypeAnnotation):
        TypeAnnotation.__init__(self, parsed=parsed)

        self.value = value

    def __str__(self) -> str:
        return f"Optional[{self.value}]"


TypeAnnotationUnion = Union[
    PrimitiveTypeAnnotation,
    OurTypeAnnotation,
    ListTypeAnnotation,
    OptionalTypeAnnotation,
]


assert_union_of_descendants_exhaustive(
    union=TypeAnnotationUnion, base_class=TypeAnnotation
)


TypeAnnotationExceptOptional = Union[
    PrimitiveTypeAnnotation,
    OurTypeAnnotation,
    ListTypeAnnotation,
]

assert_union_without_excluded(
    original_union=TypeAnnotationUnion,
    subset_union=TypeAnnotationExceptOptional,
    excluded=[OptionalTypeAnnotation],
)

AtomicTypeAnnotation = Union[PrimitiveTypeAnnotation, OurTypeAnnotation]

assert_union_without_excluded(
    original_union=TypeAnnotationUnion,
    subset_union=AtomicTypeAnnotation,
    excluded=[ListTypeAnnotation, OptionalTypeAnnotation],
)


def type_annotations_equal(
    that: TypeAnnotationUnion, other: TypeAnnotationUnion
) -> bool:
    """
    Compare two type annotations for equality.

    Two type annotations are equal if they describe the same type.
    """
    if type(that) is not type(other):
        return False

    if isinstance(that, PrimitiveTypeAnnotation):
        assert isinstance(other, PrimitiveTypeAnnotation)
        return that.a_type == other.a_type

    elif isinstance(that, OurTypeAnnotation):
        assert isinstance(other, OurTypeAnnotation)
        return that.symbol == other.symbol

    elif isinstance(that, ListTypeAnnotation):
        assert isinstance(other, ListTypeAnnotation)
        return type_annotations_equal(that.items, other.items)

    elif isinstance(that, OptionalTypeAnnotation):
        assert isinstance(other, OptionalTypeAnnotation)
        return type_annotations_equal(that.value, other.value)

    else:
        assert_never(that)

    raise AssertionError("Should not have gotten here")


def beneath_optional(
    type_annotation: TypeAnnotationUnion,
) -> TypeAnnotationExceptOptional:
    """Descend below ``Optional[...]`` to the underlying type."""
    type_anno = type_annotation
    while isinstance(type_anno, OptionalTypeAnnotation):
        type_anno = type_anno.value

    assert not isinstance(type_anno, OptionalTypeAnnotation)

    return type_anno


# region Descriptions

# NOTE (mristin, 2022-03-18):
# We take C# documentation comments as an orientation for the structure of the
# descriptions.


def find_first_field_list(
    element: docutils.nodes.Element,
) -> Optional[docutils.nodes.field_list]:
    """Find the first field list beneath the element or return None."""
    return next(element.findall(condition=docutils.nodes.field_list), None)


class SummaryRemarksDescription(DBC):
    """Represent a description with a summary and remarks."""

    #: Summary as the first line of the docstring
    summary: Final[docutils.nodes.paragraph]

    #: List of remarks following the summary in the docstring
    remarks: Final[Sequence[docutils.nodes.Element]]

    #: Original parsed description
    parsed: Final[parse.Description]

    # fmt: off
    @require(
        lambda summary:
        find_first_field_list(summary) is None,
        "Summary expected without field lists"
    )
    @require(
        lambda remarks:
        all(
            find_first_field_list(remark) is None
            for remark in remarks
        ),
        "Remarks expected without field lists"
    )
    # fmt: on
    def __init__(
        self,
        summary: docutils.nodes.paragraph,
        remarks: Sequence[docutils.nodes.Element],
        parsed: parse.Description,
    ) -> None:
        """Initialize with the given values."""
        self.summary = summary
        self.remarks = remarks
        self.parsed = parsed

    @abc.abstractmethod
    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        raise NotImplementedError()


# noinspection PyAbstractClass
class SummaryRemarksConstraintsDescription(SummaryRemarksDescription):
    """Represent a description with summary, remarks and constraints blocks."""

    #: Map constraint documentation elements by their identifiers
    constraints_by_identifier: Final[OrderedDict[str, docutils.nodes.field_body]]

    # fmt: off
    @require(
        lambda constraints_by_identifier:
        all(
            find_first_field_list(body) is None
            for body in constraints_by_identifier.values()
        ),
        "Constraint bodies expected without field lists"
    )
    # fmt: on
    def __init__(
        self,
        summary: docutils.nodes.paragraph,
        remarks: Sequence[docutils.nodes.Element],
        constraints_by_identifier: OrderedDict[str, docutils.nodes.field_body],
        parsed: parse.Description,
    ) -> None:
        """Initialize with the given values."""
        SummaryRemarksDescription.__init__(
            self, summary=summary, remarks=remarks, parsed=parsed
        )
        self.constraints_by_identifier = constraints_by_identifier


class MetaModelDescription(SummaryRemarksConstraintsDescription):
    """Represent a description of a meta-model."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class SymbolDescription(SummaryRemarksConstraintsDescription):
    """Represent a description of a symbol."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class PropertyDescription(SummaryRemarksConstraintsDescription):
    """Represent a documentation of a property."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class EnumerationLiteralDescription(SummaryRemarksDescription):
    """Represent a documentation of an enumeration literal."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class SignatureDescription(SummaryRemarksDescription):
    """Represent a documentation of a method or a function signature."""

    #: Map argument documentation by the argument names
    arguments_by_name: Final[OrderedDict[Identifier, docutils.nodes.field_body]]

    #: Documentation of the return value, if written
    returns: Final[Optional[docutils.nodes.field_body]]

    # fmt: off
    @require(
        lambda arguments_by_name:
        all(
            find_first_field_list(body) is None
            for body in arguments_by_name.values()
        ),
        "Argument descriptions expected without field lists"
    )
    @require(
        lambda returns:
        not (returns is not None)
        or find_first_field_list(returns) is None,
        "Return value description expected without field lists"
    )
    # fmt: on
    def __init__(
        self,
        summary: docutils.nodes.paragraph,
        remarks: Sequence[docutils.nodes.Element],
        arguments_by_name: OrderedDict[Identifier, docutils.nodes.field_body],
        returns: Optional[docutils.nodes.field_body],
        parsed: parse.Description,
    ) -> None:
        """Initialize with the given values."""
        SummaryRemarksDescription.__init__(
            self, summary=summary, remarks=remarks, parsed=parsed
        )
        self.arguments_by_name = arguments_by_name
        self.returns = returns

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


# endregion


class Property:
    """Represent a property of a class."""

    #: Name of the property
    name: Final[Identifier]

    #: Type annotation of the property
    type_annotation: Final[TypeAnnotationUnion]

    #: Description of the property, if any
    description: Final[Optional[PropertyDescription]]

    #: The original class where this property is specified.
    #: We stack all the properties over the ancestors, so using ``specified_for``
    #: you can distinguish between inherited properties and genuine properties of
    #: a class.
    specified_for: Final["Class"]

    #: Relation to the property from the parse stage
    parsed: Final[parse.Property]

    def __init__(
        self,
        name: Identifier,
        type_annotation: TypeAnnotationUnion,
        description: Optional[PropertyDescription],
        specified_for: "Class",
        parsed: parse.Property,
    ) -> None:
        """Initialize with the given values."""
        self.name = name
        self.type_annotation = type_annotation
        self.description = description
        self.specified_for = specified_for
        self.parsed = parsed

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class DefaultConstant:
    """Represent a constant value as a default for an argument."""

    #: The default value
    value: Final[Union[bool, int, float, str, None]]

    #: Relation to the parsed stage
    parsed: Final[parse.Default]

    def __init__(
        self, value: Union[bool, int, float, str, None], parsed: parse.Default
    ) -> None:
        """Initialize with the given values."""
        self.value = value
        self.parsed = parsed


class DefaultEnumerationLiteral:
    """Represent an enumeration literal as a default for an argument."""

    #: Related enumeration
    enumeration: Final["Enumeration"]

    #: Related enumeration literal
    literal: Final["EnumerationLiteral"]

    #: Relation to the parse stage
    parsed: Final[parse.Default]

    # fmt: off
    @require(
        lambda enumeration, literal:
        literal.name in enumeration.literals_by_name
        and enumeration.literals_by_name[literal.name] == literal
    )
    # fmt: on
    def __init__(
        self,
        enumeration: "Enumeration",
        literal: "EnumerationLiteral",
        parsed: parse.Default,
    ) -> None:
        """Initialize with the given values."""
        self.parsed = parsed
        self.enumeration = enumeration
        self.literal = literal


Default = Union[DefaultConstant, DefaultEnumerationLiteral]


class Argument:
    """Represent an argument of a method (both of an interface and of class)."""

    #: Name of the argument
    name: Final[Identifier]

    #: Type annotation of the argument
    type_annotation: Final[TypeAnnotationUnion]

    #: Default value of the argument, if any
    default: Final[Optional[Default]]

    #: Relation to the parse stage
    parsed: Final[parse.Argument]

    def __init__(
        self,
        name: Identifier,
        type_annotation: TypeAnnotationUnion,
        default: Optional[Default],
        parsed: parse.Argument,
    ) -> None:
        """Initialize with the given values."""
        self.name = name
        self.type_annotation = type_annotation
        self.default = default
        self.parsed = parsed


class Serialization:
    """Specify the general settings for serialization of an interface or a class."""

    def __init__(self, with_model_type: bool) -> None:
        """
        Initialize with the given values.

        :param with_model_type:
            if set, the serialization needs to include a discriminator.
        """
        self.with_model_type = with_model_type


class ReferenceInTheBook:
    """Represent the information indicated in the ``reference_in_the_book`` marker."""

    #: Section number
    section: Final[Tuple[int, ...]]

    #: Index in the section so that the classes can be sorted deterministically
    index: Final[int]

    #: URL Fragment of the section
    #:
    #: The literal ``#`` needs to be prepended and the fragment needs to be URL-encoded.
    fragment: Final[Optional[str]]

    def __init__(
        self, section: Tuple[int, ...], index: int, fragment: Optional[str]
    ) -> None:
        """Initialize with the given values."""
        self.section = section
        self.index = index
        self.fragment = fragment


class Invariant:
    """Represent an invariant of a class."""

    #: Human-readable description of the invariant, if any
    description: Final[Optional[str]]

    #: Understood body of the invariant
    body: Final[parse_tree.Expression]

    #: The original symbol where this invariant is specified.
    #: We stack all the invariants over the ancestors, so using ``specified_for``
    #: you can distinguish between inherited invariants and genuine invariants of
    #: a class or a constrained primitive.
    specified_for: Final[Union["ConstrainedPrimitive", "Class"]]

    #: Relation to the parse stage
    parsed: Final[parse.Invariant]

    def __init__(
        self,
        description: Optional[str],
        body: parse_tree.Expression,
        specified_for: Union["ConstrainedPrimitive", "Class"],
        parsed: parse.Invariant,
    ) -> None:
        self.description = description
        self.body = body
        self.specified_for = specified_for
        self.parsed = parsed


class Contract:
    """Represent a contract of a method."""

    #: Argument names of the contract
    args: Final[Sequence[Identifier]]

    #: Human-readable description of the contract, if any
    description: Final[Optional[str]]

    #: Understood body of the contract
    body: Final[parse_tree.Node]

    #: Relation to the parse stage
    parsed: Final[parse.Contract]

    def __init__(
        self,
        args: Sequence[Identifier],
        description: Optional[str],
        body: parse_tree.Node,
        parsed: parse.Contract,
    ) -> None:
        """Initialize with the given values."""
        self.args = args
        self.description = description
        self.body = body
        self.parsed = parsed


class Snapshot:
    """Represent a snapshot of an OLD value capture before the method execution."""

    #: Argument names of the snapshot
    args: Final[Sequence[Identifier]]

    #: Understood body of the snapshot
    body: Final[parse_tree.Node]

    #: Name of the snapshot variable
    name: Final[Identifier]

    #: Relation to parse stage
    parsed: Final[parse.Snapshot]

    def __init__(
        self,
        args: Sequence[Identifier],
        body: parse_tree.Node,
        name: Identifier,
        parsed: parse.Snapshot,
    ) -> None:
        """Initialize with the given values."""
        self.args = args
        self.body = body
        self.name = name
        self.parsed = parsed


class Contracts:
    """Represent the set of contracts for a method or a function."""

    def __init__(
        self,
        preconditions: Sequence[Contract],
        snapshots: Sequence[Snapshot],
        postconditions: Sequence[Contract],
    ) -> None:
        """Initialize with the given values."""
        self.preconditions = preconditions
        self.snapshots = snapshots
        self.postconditions = postconditions


class SignatureLike(DBC):
    """
    Represent a signature-like "something".

    This can be either a signature of a method, a method or a function.
    """

    #: Name of the signature-like
    name: Final[Identifier]

    #: Arguments of the signature-like
    arguments: Final[Sequence[Argument]]

    #: Return type of the signature-like
    returns: Final[Optional[TypeAnnotationUnion]]

    #: Description of the signature-like, if any
    description: Final[Optional[SignatureDescription]]

    #: List of contracts of the signature-like. The contracts are stacked from the
    #: ancestors.
    contracts: Final[Contracts]

    # NOTE (mristin, 2021-12-26):
    # The ``parsed`` must be optional since constructors can be synthesized without
    # being defined in the original meta-model.

    #: Relation to the parse stage
    parsed: Optional[parse.Method]

    #: Map arguments by their names
    arguments_by_name: Final[Mapping[Identifier, Argument]]

    # fmt: off
    @require(
        lambda arguments:
        all(
            arg.name != 'self'
            for arg in arguments
        ),
        "No explicit ``self`` argument in the arguments"
    )
    @require(
        lambda arguments: (
                arg_names := [arg.name for arg in arguments],
                len(arg_names) == len(set(arg_names))
        )[1],
        "Unique arguments"
    )
    @ensure(
        lambda self:
        len(self.arguments) == len(self.arguments_by_name)
        and all(
            (
                    found_argument := self.arguments_by_name.get(argument.name, None),
                    found_argument is not None and found_argument is argument
            )[1]
            for argument in self.arguments
        ),
        "Arguments and arguments-by-name consistent"
    )
    # fmt: on
    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        parsed: Optional[parse.Method],
    ) -> None:
        """Initialize with the given values."""
        self.name = name
        self.arguments = arguments
        self.returns = returns
        self.description = description
        self.contracts = contracts
        self.parsed = parsed

        self.arguments_by_name = {
            argument.name: argument for argument in self.arguments
        }

    @abc.abstractmethod
    def __repr__(self) -> str:
        # Signal that this is a pure abstract class
        raise NotImplementedError()


class Method(SignatureLike):
    """Represent a method of a class."""

    # NOTE (mristin, 2021-12-26):
    # The ``parsed`` must be optional in the parent class, ``SignatureLike``, since
    # constructors can be synthesized without being defined in the original meta-model.
    #
    # However, methods are never synthesized so we always have a clear link to the parse
    # stage.

    parsed: parse.Method

    # fmt: off
    @require(
        lambda name:
        name != "__init__",
        "Expected constructors to be handled in a special way and not as a method"
    )
    @require(
        lambda arguments, contracts:
        (
                arg_set := {arg.name for arg in arguments},
                all(
                    arg in arg_set  # pylint: disable=used-before-assignment
                    for precondition in contracts.preconditions
                    for arg in precondition.args
                    if arg != 'self'
                )
                and all(
                    arg in arg_set
                    for postcondition in contracts.postconditions
                    for arg in postcondition.args
                    if arg not in ('OLD', 'result', 'self')
                )
                and all(
                    arg in arg_set
                    for snapshot in contracts.snapshots
                    for arg in snapshot.args
                    if arg.name != 'self'
                )
        )[1],
        "All arguments of contracts defined in method arguments except ``self``"
    )
    # fmt: on
    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        parsed: parse.Method,
    ) -> None:
        """Initialize with the given values."""
        SignatureLike.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

    @abc.abstractmethod
    def __repr__(self) -> str:
        # Signal that this is a pure abstract class.
        raise NotImplementedError()


# NOTE (mristin, 2021-12-19):
# At the moment, we support only implementation-specific methods. However, we anticipate
# that we will try to understand the methods in the very near future so we already
# prepare the class hierarchy for it.


class ImplementationSpecificMethod(Method):
    """Represent an implementation-specific method of a class."""

    # NOTE (mristin, 2021-12-26):
    # The ``parsed`` must be optional in the parent class, ``SignatureLike``, since
    # constructors can be synthesized without being defined in the original meta-model.
    #
    # However, methods are never synthesized so we always have a clear link to the parse
    # stage here.

    #: Relation to parse stage
    parsed: parse.Method

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class UnderstoodMethod(Method):
    """Represent a method of a class which we could understand."""

    #: Understood syntax tree of the method's body
    body: Final[Sequence[parse_tree.Node]]

    # NOTE (mristin, 2021-12-26):
    # The ``parsed`` must be optional in the parent class, ``SignatureLike``, since
    # constructors can be synthesized without being defined in the original meta-model.
    #
    # However, methods are never synthesized so we always have a clear link to the parse
    # stage here.

    #: Relation to parse stage
    parsed: parse.Method

    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        body: Sequence[parse_tree.Node],
        parsed: parse.Method,
    ) -> None:
        """Initialize with the given values."""
        Method.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

        self.body = body

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class Constructor(SignatureLike):
    """
    Represent an understood constructor of a class stacked.

    The constructor is expected to be stacked from the class and all the ancestors.
    """

    #: Interpreted statements of the constructor, stacked over all the ancestors
    statements: Final[Sequence[construction.AssignArgument]]

    #: If set, the constructor is implementation-specific and we need to provide
    #: a snippet for it.
    is_implementation_specific: Final[bool]

    def __init__(
        self,
        is_implementation_specific: bool,
        arguments: Sequence[Argument],
        contracts: Contracts,
        description: Optional[SignatureDescription],
        statements: Sequence[construction.AssignArgument],
        parsed: Optional[parse.Method],
    ) -> None:
        SignatureLike.__init__(
            self,
            name=Identifier("__init__"),
            arguments=arguments,
            returns=None,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

        self.is_implementation_specific = is_implementation_specific

        # The calls to the super constructors must be in-lined before.
        self.statements = statements

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class EnumerationLiteral:
    """Represent a single enumeration literal."""

    def __init__(
        self,
        name: Identifier,
        value: str,
        description: Optional[EnumerationLiteralDescription],
        parsed: parse.EnumerationLiteral,
    ) -> None:
        self.name = name
        self.value = value
        self.description = description
        self.parsed = parsed


# fmt: off
@invariant(
    lambda self:
    all(
        literal == self.literals_by_name[literal.name]
        for literal in self.literals
    ),
    "Literal map consistent on name"
)
@invariant(
    lambda self:
    sorted(map(id, self.literals_by_name.values())) == sorted(map(id, self.literals)),
    "Literal map complete"
)
# fmt: on
class Enumeration:
    """Represent an enumeration."""

    #: Name of the enumeration
    name: Final[Identifier]

    #: Literals associated with the enumeration
    literals: Final[Sequence[EnumerationLiteral]]

    #: List which enumerations *this* enumeration is a superset of;
    #: this is akin to inheritance for enumerations
    is_superset_of: Final[Sequence["Enumeration"]]

    #: Reference to the original specs
    reference_in_the_book: Final[Optional[ReferenceInTheBook]]

    #: Description of the enumeration, if any
    description: Final[Optional[SymbolDescription]]

    #: Map literals by their identifiers
    literals_by_name: Final[Mapping[str, EnumerationLiteral]]

    #: Collect IDs (with :py:func:`id`) of the literal objects in a set
    literal_id_set: Final[FrozenSet[int]]

    def __init__(
        self,
        name: Identifier,
        literals: Sequence[EnumerationLiteral],
        is_superset_of: Sequence["Enumeration"],
        reference_in_the_book: Optional[ReferenceInTheBook],
        description: Optional[SymbolDescription],
        parsed: parse.Enumeration,
    ) -> None:
        self.name = name
        self.literals = literals
        self.is_superset_of = is_superset_of
        self.reference_in_the_book = reference_in_the_book
        self.description = description
        self.parsed = parsed

        self.literals_by_name: Mapping[str, EnumerationLiteral] = {
            literal.name: literal for literal in self.literals
        }

        self.literal_id_set = frozenset(id(literal) for literal in literals)

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class ConstrainedPrimitive:
    """Represent a primitive type constrained by one or more invariants."""

    #: Name of the class
    name: Final[Identifier]

    # region Inheritances

    # NOTE (mristin, 2021-12-24):
    # We have to decorate inheritances with ``@property`` so that the client code is
    # forced to use ``_set_inheritances``.

    _inheritances: Sequence["ConstrainedPrimitive"]

    @property
    def inheritances(self) -> Sequence["ConstrainedPrimitive"]:
        """Return direct parents that this class inherits from."""
        return self._inheritances

    _inheritance_id_set: FrozenSet[int]

    @property
    def inheritance_id_set(self) -> FrozenSet[int]:
        """Collect IDs (with :py:func:`id`) of the inheritance objects in a set."""
        return self._inheritance_id_set

    # endregion

    # region Descendants

    # NOTE (mristin, 2021-12-24):
    # We have to decorate ``descendant_id_set`` with
    # ``@property`` so that the translation code is forced to use
    # ``_set_descendants``.

    _descendant_id_set: FrozenSet[int]

    @property
    def descendant_id_set(self) -> FrozenSet[int]:
        """List the IDs (as in Python's ``id`` built-in) of the descendants."""
        return self._descendant_id_set

    # endregion

    #: Which primitive type is constrained
    constrainee: PrimitiveType

    #: If set, this class is implementation-specific and we need to provide a snippet
    #: for each implementation target
    is_implementation_specific: Final[bool]

    # region Invariants

    # NOTE (mristin, 2022-03-19):
    # We have to decorate invariants with ``@property`` so that the translation code
    # is forced to use ``_set_invariants``.

    _invariants: Sequence[Invariant]

    @property
    def invariants(self) -> Sequence[Invariant]:
        """List invariants of the class."""
        return self._invariants

    _invariant_id_set: FrozenSet[int]

    @property
    def invariant_id_set(self) -> FrozenSet[int]:
        """Collect IDs (with :py:func:`id`) of the invariant objects in a set."""
        return self._invariant_id_set

    # endregion

    #: Reference to the original specs
    reference_in_the_book: Final[Optional[ReferenceInTheBook]]

    #: Description of the class
    description: Final[Optional[SymbolDescription]]

    #: Relation to the class from the parse stage
    parsed: parse.Class

    # fmt: off
    @require(
        lambda parsed: len(parsed.methods) == 0,
        "No methods expected in the constrained primitive type"
    )
    @require(
        lambda parsed: len(parsed.properties) == 0,
        "No properties expected in the constrained primitive type"
    )
    @require(
        lambda constrainee, inheritances:
        all(
            inheritance.constrainee == constrainee
            for inheritance in inheritances
        ),
        "Constrainee consistent with ancestors"
    )
    @require(
        lambda constrainee, descendants:
        all(
            descendant.constrainee == constrainee
            for descendant in descendants
        ),
        "Constrainee consistent with descendants"
    )
    # fmt: on
    def __init__(
        self,
        name: Identifier,
        inheritances: Sequence["ConstrainedPrimitive"],
        descendants: Sequence["ConstrainedPrimitive"],
        constrainee: PrimitiveType,
        is_implementation_specific: bool,
        invariants: Sequence[Invariant],
        reference_in_the_book: Optional[ReferenceInTheBook],
        description: Optional[SymbolDescription],
        parsed: parse.Class,
    ) -> None:
        self.name = name
        self._set_inheritances(inheritances)
        self._set_descendants(descendants)
        self.constrainee = constrainee
        self.is_implementation_specific = is_implementation_specific
        self._set_invariants(invariants)
        self.reference_in_the_book = reference_in_the_book
        self.description = description
        self.parsed = parsed

    def _set_descendants(self, descendants: Sequence["ConstrainedPrimitive"]) -> None:
        """
        Set the descendants in the constrained primitive.

        This method is expected to be called only during the translation phase.
        """
        self._descendant_id_set = frozenset(
            id(descendant) for descendant in descendants
        )

    def _set_invariants(self, invariants: Sequence[Invariant]) -> None:
        """
        Set the invariants in the class.

        This method is expected to be called only during the translation phase.
        """
        self._invariants = invariants
        self._invariant_id_set = frozenset(id(invariant) for invariant in invariants)

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )

    # fmt: off
    @require(
        lambda inheritances:
        len(inheritances) == len(set(inheritance.name for inheritance in inheritances)),
        "No duplicate inheritances"
    )
    # fmt: on
    def _set_inheritances(self, inheritances: Sequence["ConstrainedPrimitive"]) -> None:
        """
        Set the inheritances in the class.

        This method is expected to be called only during the translation phase.
        """
        self._inheritances = inheritances

        self._inheritance_id_set = frozenset(
            id(inheritance) for inheritance in self._inheritances
        )


class Class(DBC):
    """Represent an abstract or a concrete class."""

    #: Name of the class
    name: Final[Identifier]

    # region Inheritances

    # NOTE (mristin, 2021-12-24):
    # We have to decorate inheritances with ``@property`` so that the translation code
    # is forced to use ``_set_inheritances``.

    _inheritances: Sequence["ClassUnion"]

    @property
    def inheritances(self) -> Sequence["ClassUnion"]:
        """Return direct parents that this class inherits from."""
        return self._inheritances

    _inheritance_id_set: FrozenSet[int]

    @property
    def inheritance_id_set(self) -> FrozenSet[int]:
        """Collect IDs (with :py:func:`id`) of the inheritance objects in a set."""
        return self._inheritance_id_set

    # endregion

    #: If set, this class is implementation-specific and we need to provide a snippet
    #: for each implementation target
    is_implementation_specific: Final[bool]

    #: Interface of the class. If it is a concrete class with no descendants, there is
    #: no interface available.
    interface: Optional["Interface"]

    # region Descendants

    # NOTE (mristin, 2021-12-24):
    # We have to decorate ``descendant_id_set`` and ``concrete_descendants`` with
    # ``@property`` so that the translation code is forced to use
    # ``_set_descendants``.

    _descendant_id_set: FrozenSet[int]

    _concrete_descendants: Sequence["ConcreteClass"]

    @property
    def descendant_id_set(self) -> FrozenSet[int]:
        """List the IDs (as in Python's ``id`` built-in) of the descendants."""
        return self._descendant_id_set

    @property
    def concrete_descendants(self) -> Sequence["ConcreteClass"]:
        """List descendants of this class which are concrete classes."""
        return self._concrete_descendants

    # endregion

    # region Properties

    # NOTE (mristin, 2022-03-19):
    # We have to decorate properties with ``@property`` so that the translation code
    # is forced to use ``_set_properties``.

    _properties: Sequence[Property]

    @property
    def properties(self) -> Sequence[Property]:
        """Return list of properties of the class."""
        return self._properties

    _properties_by_name: Mapping[Identifier, Property]

    @property
    def properties_by_name(self) -> Mapping[Identifier, Property]:
        """Map all properties by their names."""
        return self._properties_by_name

    _property_id_set: FrozenSet[int]

    @property
    def property_id_set(self) -> FrozenSet[int]:
        """Collect IDs (with :py:func:`id`) of the property objects in a set."""
        return self._property_id_set

    # endregion

    # region Methods

    # NOTE (mristin, 2022-03-19):
    # We have to decorate methods with ``@property`` so that the translation code
    # is forced to use ``_set_methods``.

    _methods: Sequence["MethodUnion"]

    @property
    def methods(self) -> Sequence["MethodUnion"]:
        """
        List methods of the class.

        The methods are strictly non-static and non-class (in the Python sense of
        the terms).
        """
        return self._methods

    _methods_by_name: Mapping[Identifier, "MethodUnion"]

    @property
    def methods_by_name(self) -> Mapping[Identifier, "MethodUnion"]:
        """Map all methods by their names."""
        return self._methods_by_name

    # endregion

    #: Constructor specification of the class
    constructor: Final[Constructor]

    # region Invariants

    # NOTE (mristin, 2022-03-19):
    # We have to decorate invariants with ``@property`` so that the translation code
    # is forced to use ``_set_invariants``.

    _invariants: Sequence[Invariant]

    @property
    def invariants(self) -> Sequence[Invariant]:
        """List invariants of the class."""
        return self._invariants

    _invariant_id_set: FrozenSet[int]

    @property
    def invariant_id_set(self) -> FrozenSet[int]:
        """Collect IDs (with :py:func:`id`) of the invariant objects in a set."""
        return self._invariant_id_set

    # endregion

    #: Particular serialization settings for this class
    serialization: Final[Serialization]

    #: Reference to the original specs
    reference_in_the_book: Final[Optional[ReferenceInTheBook]]

    #: Description of the class
    description: Final[Optional[SymbolDescription]]

    #: Relation to the class from the parse stage
    parsed: Final[parse.Class]

    def __init__(
        self,
        name: Identifier,
        inheritances: Sequence["ClassUnion"],
        interface: Optional["Interface"],
        descendants: Sequence["ClassUnion"],
        is_implementation_specific: bool,
        properties: Sequence[Property],
        methods: Sequence["MethodUnion"],
        constructor: Constructor,
        invariants: Sequence[Invariant],
        serialization: Serialization,
        reference_in_the_book: Optional[ReferenceInTheBook],
        description: Optional[SymbolDescription],
        parsed: parse.Class,
    ) -> None:
        """Initialize with the given values."""
        self.name = name
        self._set_inheritances(inheritances)
        self.interface = interface
        self._set_descendants(descendants)
        self.is_implementation_specific = is_implementation_specific
        self._set_properties(properties)
        self._set_methods(methods)
        self.constructor = constructor
        self._set_invariants(invariants)
        self.serialization = serialization
        self.reference_in_the_book = reference_in_the_book
        self.description = description
        self.parsed = parsed

    # fmt: off
    @require(
        lambda inheritances:
        len(inheritances) == len(set(inheritance.name for inheritance in inheritances)),
        "No duplicate inheritances"
    )
    # fmt: on
    def _set_inheritances(self, inheritances: Sequence["ClassUnion"]) -> None:
        """
        Set the inheritances in the class.

        This method is expected to be called only during the translation phase.
        """
        self._inheritances = inheritances

        self._inheritance_id_set = frozenset(
            id(inheritance) for inheritance in self._inheritances
        )

    def _set_descendants(self, descendants: Sequence["ClassUnion"]) -> None:
        """
        Set the descendants and the concrete descendants in the class.

        This method is expected to be called only during the translation phase.
        """
        self._descendant_id_set = frozenset(
            id(descendant) for descendant in descendants
        )

        self._concrete_descendants = [
            descendant
            for descendant in descendants
            if isinstance(descendant, ConcreteClass)
        ]

    # fmt: off
    @require(
        lambda properties:
        len(properties) == len(set(prop.name for prop in properties)),
        "No duplicate properties"
    )
    # fmt: on
    def _set_properties(self, properties: Sequence[Property]) -> None:
        """
        Set the properties in the class.

        This method is expected to be called only during the translation phase.
        """
        self._properties = properties
        self._properties_by_name = {prop.name: prop for prop in properties}
        self._property_id_set = frozenset(id(prop) for prop in properties)

    # fmt: off
    @require(
        lambda methods:
        len(methods) == len(set(method.name for method in methods)),
        "No duplicate methods"
    )
    # fmt: on
    def _set_methods(self, methods: Sequence["MethodUnion"]) -> None:
        """
        Set the methods in the class.

        This method is expected to be called only during the translation phase.
        """
        self._methods = methods
        self._methods_by_name = {method.name: method for method in methods}

    def _set_invariants(self, invariants: Sequence[Invariant]) -> None:
        """
        Set the invariants in the class.

        This method is expected to be called only during the translation phase.
        """
        self._invariants = invariants
        self._invariant_id_set = frozenset(id(invariant) for invariant in invariants)

    @abc.abstractmethod
    def __repr__(self) -> str:
        # Signal that this is a purely abstract class.
        raise NotImplementedError()


class ConcreteClass(Class):
    """Represent a class that can be instantiated."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class AbstractClass(Class):
    """Represent a class that is purely abstract and can not be instantiated."""

    #: Interface of the class. All abstract classes have an interface as opposed to
    #: concrete classes, which only have an interface if there are descendants.
    interface: "Interface"

    # We need to override the constructor because the ``interface`` is required.
    def __init__(
        self,
        name: Identifier,
        inheritances: Sequence["ClassUnion"],
        interface: "Interface",
        descendants: Sequence["ClassUnion"],
        is_implementation_specific: bool,
        properties: Sequence[Property],
        methods: Sequence["MethodUnion"],
        constructor: Constructor,
        invariants: Sequence[Invariant],
        serialization: Serialization,
        reference_in_the_book: Optional[ReferenceInTheBook],
        description: Optional[SymbolDescription],
        parsed: parse.Class,
    ) -> None:
        """Initialize with the given values."""
        Class.__init__(
            self,
            name=name,
            inheritances=inheritances,
            interface=interface,
            descendants=descendants,
            is_implementation_specific=is_implementation_specific,
            properties=properties,
            methods=methods,
            constructor=constructor,
            invariants=invariants,
            serialization=serialization,
            reference_in_the_book=reference_in_the_book,
            description=description,
            parsed=parsed,
        )

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class Verification(SignatureLike):
    """Represent a verification function defined in the meta-model."""

    parsed: parse.Method

    # fmt: off
    @require(
        lambda arguments, contracts:
        (
                arg_set := {arg.name for arg in arguments},
                all(
                    arg in arg_set  # pylint: disable=used-before-assignment
                    for precondition in contracts.preconditions
                    for arg in precondition.args
                )
                and all(
                    arg in arg_set
                    for postcondition in contracts.postconditions
                    for arg in postcondition.args
                    if arg not in ('OLD', 'result')
                )
                and all(
                    arg in arg_set
                    for snapshot in contracts.snapshots
                    for arg in snapshot.args
                )
        )[1],
        "All arguments of contracts defined in function arguments"
    )
    # fmt: on
    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        parsed: parse.Method,
    ) -> None:
        """Initialize with the given values."""
        SignatureLike.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

    @abc.abstractmethod
    def __repr__(self) -> str:
        # Signal that this is a pure abstract class.
        raise NotImplementedError()


class ImplementationSpecificVerification(Verification):
    """Represent an implementation-specific verification function."""

    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        parsed: parse.Method,
    ) -> None:
        """Initialize with the given values."""
        Verification.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class PatternVerification(Verification):
    """
    Represent a function that checks a string against a regular expression.

    There is expected to be a single string argument (the text to be matched).
    The function is expected to return a boolean.
    """

    #: Method as we understood it in the parse stage
    parsed: parse.UnderstoodMethod

    #: Pattern, *i.e.* the regular expression, that the function checks against
    pattern: Final[str]

    # fmt: off
    @require(
        lambda arguments:
        len(arguments) == 1
        and isinstance(arguments[0].type_annotation, PrimitiveTypeAnnotation)
        and arguments[0].type_annotation.a_type == PrimitiveType.STR,
        "There is a single string argument"
    )
    @require(
        lambda returns:
        (returns is not None)
        and isinstance(returns, PrimitiveTypeAnnotation)
        and returns.a_type == PrimitiveType.BOOL
    )
    # fmt: on
    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        pattern: str,
        parsed: parse.UnderstoodMethod,
    ) -> None:
        """Initialize with the given values."""
        Verification.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

        self.pattern = pattern

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class Signature(SignatureLike):
    """Represent a signature of a method in an interface."""

    def __init__(
        self,
        name: Identifier,
        arguments: Sequence[Argument],
        returns: Optional[TypeAnnotationUnion],
        description: Optional[SignatureDescription],
        contracts: Contracts,
        parsed: parse.Method,
    ) -> None:
        """
        Initialize with the given values.

        The ``parsed`` refers to the method of the abstract or concrete class that
        defines the interface. Mind that we do not introduce interfaces as a concept
        in the meta-model.
        """
        SignatureLike.__init__(
            self,
            name=name,
            arguments=arguments,
            returns=returns,
            description=description,
            contracts=contracts,
            parsed=parsed,
        )

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


class Interface:
    """
    Represent an interface of some of the abstract and/or concrete classes.

    Mind that the concept of interfaces is *not* used in the meta-model. We introduce
    it at the intermediate stage to facilitate generation of the code, especially for
    targets where multiple inheritance is not supported.
    """

    #: Class which this interface is based on
    base: Final[Class]

    #: Name of the interface
    name: Final[Identifier]

    inheritances: Final[Sequence["Interface"]]

    #: List of concrete classes that implement this interface
    implementers: Sequence["ConcreteClass"]

    #: List of properties assumed by the interface
    properties: Final[Sequence[Property]]

    #: List of method signatures assumed by the interface
    signatures: Final[Sequence[Signature]]

    #: Description of the interface, taken from class
    description: Final[Optional[SymbolDescription]]

    #: Relation to the class from the parse stage
    parsed: Final[parse.Class]

    #: Map all properties by their identifiers to the corresponding objects
    properties_by_name: Final[Mapping[Identifier, Property]]

    #: Collect IDs (with :py:func:`id`) of the property objects in a set
    property_id_set: Final[FrozenSet[int]]

    def __init__(
        self,
        base: Class,
        inheritances: Sequence["Interface"],
    ) -> None:
        """Initialize with the given values."""
        self.base = base

        self.name = base.name
        self.inheritances = inheritances

        implementers = list(base.concrete_descendants)

        if isinstance(base, ConcreteClass):
            implementers.append(base)

        self.implementers = implementers

        self.properties = base.properties

        self.signatures = [
            Signature(
                name=method.name,
                arguments=method.arguments,
                returns=method.returns,
                description=method.description,
                contracts=method.contracts,
                parsed=method.parsed,
            )
            for method in base.methods
        ]

        self.description = base.description
        self.parsed = base.parsed

        self.properties_by_name: Mapping[Identifier, Property] = {
            prop.name: prop for prop in self.properties
        }

        self.property_id_set = frozenset(id(prop) for prop in self.properties)

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return (
            f"<{_MODULE_NAME}.{self.__class__.__name__} {self.name} at 0x{id(self):x}>"
        )


T = TypeVar("T")


class MetaModel:
    """Collect information about the underlying meta-model."""

    #: Description of the meta-model extracted from the docstring
    description: Final[Optional[MetaModelDescription]]

    #: Specify the URL of the book that the meta-model is based on
    book_url: Final[str]

    #: Specify the version of the book that the meta-model is based on
    book_version: Final[str]

    def __init__(
        self,
        book_url: str,
        book_version: str,
        description: Optional[MetaModelDescription],
    ) -> None:
        self.book_url = book_url
        self.book_version = book_version
        self.description = description


class SymbolTable:
    """Represent all the symbols of the intermediate representation."""

    #: List of all symbols that we need for the code generation
    symbols: Final[Sequence["Symbol"]]

    #: List of all the symbols, topologically sorted by inheritance
    symbols_topologically_sorted: Final[Sequence["Symbol"]]

    #: List of all functions used in the verification
    verification_functions: Final[Sequence["VerificationUnion"]]

    #: Map verification functions by their name
    verification_functions_by_name: Final[Mapping[Identifier, "VerificationUnion"]]

    #: Additional information about the source meta-model
    meta_model: Final[MetaModel]

    _name_to_symbol: Final[Mapping[Identifier, "Symbol"]]

    # fmt: off
    @require(
        lambda symbols: (
                names := [symbol.name for symbol in symbols],
                len(names) == len(set(names)),
        )[1],
        "Symbol names unique",
    )
    @require(
        lambda symbols, symbols_topologically_sorted:
        set(
            id(symbol)
            for symbol in symbols
            if not isinstance(symbol, Enumeration)
        )
        == set(id(symbol) for symbol in symbols_topologically_sorted),
        "Only maybe the order differs between the symbols and "
        "symbols_topologically_sorted"
    )
    @ensure(
        lambda self:
        all(
            self.verification_functions_by_name[func.name] is func
            for func in self.verification_functions
        )
        and len(self.verification_functions_by_name) == len(
            self.verification_functions),
        "The verification functions and their mapping by name are consistent"
    )
    @ensure(
        lambda self:
        all(
            (
                    found_symbol := self.find(symbol.name),
                    found_symbol is not None and found_symbol is symbol
            )[1]
            for symbol in self.symbols
        ),
        "Finding symbols is consistent with ``symbols``"
    )
    # fmt: on
    def __init__(
        self,
        symbols: Sequence["Symbol"],
        symbols_topologically_sorted: Sequence["SymbolExceptEnumeration"],
        verification_functions: Sequence["VerificationUnion"],
        meta_model: MetaModel,
    ) -> None:
        """Initialize with the given values and map symbols to name."""
        self.symbols = symbols
        self.symbols_topologically_sorted = symbols_topologically_sorted
        self.verification_functions = verification_functions
        self.meta_model = meta_model

        self.verification_functions_by_name = {
            func.name: func for func in self.verification_functions
        }

        self._name_to_symbol = {symbol.name: symbol for symbol in symbols}

    def find(self, name: Identifier) -> Optional["Symbol"]:
        """Find the symbol with the given ``name``."""
        return self._name_to_symbol.get(name, None)

    def must_find(self, name: Identifier) -> "Symbol":
        """
        Find the symbol with the given ``name``.

        :raise: :py:class:`KeyError` if the ``name`` is not in the table.
        """
        result = self.find(name)
        if result is None:
            raise KeyError(
                f"Could not find the symbol with the name {name} "
                f"in the symbol table"
            )

        return result


def map_descendability(
    type_annotation: TypeAnnotationUnion,
) -> MutableMapping[TypeAnnotationUnion, bool]:
    """
    Map the type annotation recursively by the descendability.

    The descendability means that the type annotation references an interface
    or a class *or* that it is a subscripted type annotation which subscribes one or
    more classes of the meta-model.

    Constrained primitives are considered primitives and thus non-descendable.

    The mapping is a form of caching. Otherwise, the time complexity would be quadratic
    if we queried at each type annotation subscript.
    """
    mapping = dict()  # type: MutableMapping[TypeAnnotationUnion, bool]

    def recurse(a_type_annotation: TypeAnnotationUnion) -> bool:
        """Recursively iterate over subscripted type annotations."""
        if isinstance(a_type_annotation, PrimitiveTypeAnnotation):
            mapping[a_type_annotation] = False
            return False

        elif isinstance(a_type_annotation, OurTypeAnnotation):
            result = None  # type: Optional[bool]
            if isinstance(a_type_annotation.symbol, Enumeration):
                result = False
            elif isinstance(a_type_annotation.symbol, ConstrainedPrimitive):
                result = False
            elif isinstance(a_type_annotation.symbol, Class):
                result = True
            else:
                assert_never(a_type_annotation.symbol)

            assert result is not None
            mapping[a_type_annotation] = result
            return result

        elif isinstance(a_type_annotation, ListTypeAnnotation):
            result = recurse(a_type_annotation=a_type_annotation.items)
            mapping[a_type_annotation] = result
            return result

        elif isinstance(a_type_annotation, OptionalTypeAnnotation):
            result = recurse(a_type_annotation=a_type_annotation.value)
            mapping[a_type_annotation] = result
            return result

        else:
            assert_never(a_type_annotation)

        raise AssertionError("Should not have gotten here")

    _ = recurse(a_type_annotation=type_annotation)

    return mapping


class _ConstructorArgumentOfClass:
    """Represent a constructor argument with its corresponding class."""

    def __init__(self, arg: Argument, cls: Class) -> None:
        """Initialize with the given values."""
        self.arg = arg
        self.cls = cls


def collect_ids_of_symbols_in_properties(symbol_table: SymbolTable) -> Set[int]:
    """
    Collect the IDs of the symbols occurring in type annotations of the properties.

    The IDs refer to IDs of the Python objects in this context.
    """
    result = set()  # type: Set[int]
    for symbol in symbol_table.symbols:
        if isinstance(symbol, Enumeration):
            continue

        elif isinstance(symbol, ConstrainedPrimitive):
            continue

        elif isinstance(symbol, Class):
            for prop in symbol.properties:
                type_anno = prop.type_annotation

                old_type_anno = None  # type: Optional[TypeAnnotation]
                while True:
                    if isinstance(type_anno, OptionalTypeAnnotation):
                        # noinspection PyUnresolvedReferences
                        type_anno = type_anno.value
                    elif isinstance(type_anno, ListTypeAnnotation):
                        type_anno = type_anno.items
                    elif isinstance(type_anno, PrimitiveTypeAnnotation):
                        break
                    elif isinstance(type_anno, OurTypeAnnotation):
                        result.add(id(type_anno.symbol))
                        break
                    else:
                        assert_never(type_anno)

                    assert old_type_anno is not type_anno, "Loop invariant"
                    old_type_anno = type_anno
        else:
            assert_never(symbol)

    return result


DescriptionUnion = Union[
    MetaModelDescription,
    SymbolDescription,
    PropertyDescription,
    EnumerationLiteralDescription,
    SignatureDescription,
]
assert_union_of_descendants_exhaustive(
    union=DescriptionUnion, base_class=SummaryRemarksDescription
)

ClassUnion = Union[AbstractClass, ConcreteClass]
assert_union_of_descendants_exhaustive(union=ClassUnion, base_class=Class)

MethodUnion = Union[UnderstoodMethod, ImplementationSpecificMethod]
assert_union_of_descendants_exhaustive(union=MethodUnion, base_class=Method)

Symbol = Union[Enumeration, ConstrainedPrimitive, ClassUnion]

SymbolExceptEnumeration = Union[ConstrainedPrimitive, ClassUnion]
assert_union_without_excluded(
    original_union=Symbol, subset_union=SymbolExceptEnumeration, excluded=[Enumeration]
)

VerificationUnion = Union[ImplementationSpecificVerification, PatternVerification]
assert_union_of_descendants_exhaustive(union=VerificationUnion, base_class=Verification)



"""Render descriptions to C# documentation comments."""
import io
import textwrap
import xml.sax.saxutils
from typing import Tuple, Optional, List

import docutils.nodes
import docutils.parsers.rst.roles
import docutils.utils
from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Error, assert_never, Identifier
from aas_core_codegen.csharp import (
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import INDENT as I
from aas_core_codegen.intermediate import (
    doc as intermediate_doc,
    rendering as intermediate_rendering,
)


class _ElementRenderer(intermediate_rendering.DocutilsElementTransformer[str]):
    """Render descriptions as C# docstring XML."""

    def transform_text(
        self, element: docutils.nodes.Text
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return xml.sax.saxutils.escape(element.astext()), None

    def transform_symbol_reference_in_doc(
        self, element: intermediate_doc.SymbolReference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        name = None  # type: Optional[str]

        if isinstance(element.symbol, intermediate.Enumeration):
            name = csharp_naming.enum_name(element.symbol.name)

        elif isinstance(element.symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2021-12-17):
            # We do not generate a class for constrained primitives, but we
            # leave it as class name, as that is what we used for ``Verify*`` function.
            name = csharp_naming.class_name(element.symbol.name)

        elif isinstance(element.symbol, intermediate.Class):
            if isinstance(element.symbol, intermediate.AbstractClass):
                # NOTE (mristin, 2021-12-25):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                name = csharp_naming.interface_name(element.symbol.name)

            elif isinstance(element.symbol, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.
                name = csharp_naming.class_name(element.symbol.name)

            else:
                assert_never(element.symbol)

        else:
            assert_never(element.symbol)

        assert name is not None
        return f"<see cref={xml.sax.saxutils.quoteattr(name)} />", None

    def transform_attribute_reference_in_doc(
        self, element: intermediate_doc.AttributeReference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        cref = None  # type: Optional[str]

        if isinstance(element.reference, intermediate_doc.PropertyReference):
            symbol_name = None  # type: Optional[str]

            if isinstance(element.reference.cls, intermediate.AbstractClass):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                symbol_name = csharp_naming.interface_name(element.reference.cls.name)
            elif isinstance(element.reference.cls, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.

                symbol_name = csharp_naming.class_name(element.reference.cls.name)
            else:
                assert_never(element.reference.cls)

            prop_name = csharp_naming.property_name(element.reference.prop.name)

            assert symbol_name is not None
            cref = f"{symbol_name}.{prop_name}"
        elif isinstance(
            element.reference, intermediate_doc.EnumerationLiteralReference
        ):
            symbol_name = csharp_naming.enum_name(element.reference.symbol.name)
            literal_name = csharp_naming.enum_literal_name(
                element.reference.literal.name
            )

            cref = f"{symbol_name}.{literal_name}"
        else:
            assert_never(element.reference)

        assert cref is not None
        return f"<see cref={xml.sax.saxutils.quoteattr(cref)} />", None

    def transform_argument_reference_in_doc(
        self, element: intermediate_doc.ArgumentReference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        arg_name = csharp_naming.argument_name(Identifier(element.reference))
        return f"<paramref name={xml.sax.saxutils.quoteattr(arg_name)} />", None

    def transform_constraint_reference_in_doc(
        self, element: intermediate_doc.ConstraintReference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return f"Constraint {element.reference}", None

    def transform_literal(
        self, element: docutils.nodes.literal
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return f"<c>{xml.sax.saxutils.escape(element.astext())}</c>", None

    def transform_paragraph(
        self, element: docutils.nodes.paragraph
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "".join(parts), None

    def transform_emphasis(
        self, element: docutils.nodes.emphasis
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "<em>{}</em>".format("".join(parts)), None

    def transform_list_item(
        self, element: docutils.nodes.list_item
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = []  # type: List[str]
        errors = []  # type: List[str]

        for child in element.children:
            text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert text is not None
                parts.append(text)

        if len(errors) > 0:
            return None, errors

        return "<li>{}</li>".format("".join(parts)), None

    def transform_bullet_list(
        self, element: docutils.nodes.bullet_list
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = ["<ul>\n"]
        errors = []  # type: List[str]

        for child in element.children:
            text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert text is not None
                parts.append(f"{text}\n")

        if len(errors) > 0:
            return None, errors

        parts.append("</ul>")

        return "".join(parts), None

    def transform_note(
        self, element: docutils.nodes.note
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = []  # type: List[str]
        errors = []  # type: List[str]

        for child in element.children:
            text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert text is not None
                parts.append(text)

        if len(errors) > 0:
            return None, errors

        return "".join(parts), None

    def transform_reference(
        self, element: docutils.nodes.reference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        parts = []  # type: List[str]
        errors = []  # type: List[str]

        for child in element.children:
            text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert text is not None
                parts.append(text)

        if len(errors) > 0:
            return None, errors

        return "".join(parts), None

    def _transform_children_joined_with_double_new_line(
        self, element: docutils.nodes.Element
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        """Transform the ``element``'s children and join them with a double new-line."""
        if len(element.children) == 0:
            return "", None

        if len(element.children) == 1:
            return self.transform(element.children[0])

        parts = []  # type: List[str]
        errors = []  # type: List[str]

        for child in element.children:
            part, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert part is not None
                parts.append(part)

        if len(errors) > 0:
            return None, errors

        return "\n\n".join(parts), None

    def transform_field_body(
        self, element: docutils.nodes.field_body
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._transform_children_joined_with_double_new_line(element=element)

    def transform_document(
        self, element: docutils.nodes.field_body
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._transform_children_joined_with_double_new_line(element=element)


@require(lambda line: "\n" not in line)
def _slash_slash_slash_line(line: str) -> str:
    """Prepend ``///`` to the ``line``."""
    if len(line) == 0:
        return "///"

    return f"/// {line}"


def _generate_summary_remarks(
    description: intermediate.SummaryRemarksDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks-constraints."""
    errors = []  # type: List[Error]
    renderer = _ElementRenderer()

    summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )

    remarks = []  # type: List[str]
    for remark in description.remarks:
        remark, remark_errors = renderer.transform(remark)
        if remark_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark is not None
            remarks.append(remark)

    if len(errors) > 0:
        return None, errors

    assert summary is not None

    # Don't use textwrap.dedent to preserve the formatting

    blocks = [
        Stripped(
            f"""\
<summary>
{summary}
</summary>"""
        )
    ]

    if len(remarks) > 0:
        remarks_joined = "\n\n".join(remarks)
        blocks.append(
            Stripped(
                f"""\
<remarks>
{remarks_joined}
</remarks>"""
            )
        )

    commented_lines = [
        _slash_slash_slash_line(line) for block in blocks for line in block.splitlines()
    ]

    return Stripped("\n".join(commented_lines)), None


def _generate_summary_remarks_constraints(
    description: intermediate.SummaryRemarksConstraintsDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks-constraints."""
    errors = []  # type: List[Error]
    renderer = _ElementRenderer()

    summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )

    remarks = []  # type: List[str]
    for remark in description.remarks:
        remark, remark_errors = renderer.transform(remark)
        if remark_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark is not None
            remarks.append(remark)

    constraints = []  # type: List[str]
    for identifier, body_element in description.constraints_by_identifier.items():
        body, body_errors = renderer.transform(body_element)
        if body_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in body_errors
            )
        else:
            assert body is not None

            constraints.append(
                f"Constraint {xml.sax.saxutils.escape(identifier)}:\n{body}"
            )

    if len(errors) > 0:
        return None, errors

    assert summary is not None

    # Don't use textwrap.dedent to preserve the formatting

    blocks = [
        Stripped(
            f"""\
<summary>
{summary}
</summary>"""
        )
    ]

    if len(constraints) > 0:
        constraints_writer = io.StringIO()
        constraints_writer.write("Constraints:\n<ul>\n")
        for constraint in constraints:
            constraints_writer.write(textwrap.indent(f"<li>\n{constraint}\n</li>\n", I))
        constraints_writer.write("</ul>")
        remarks.append(constraints_writer.getvalue())

    if len(remarks) > 0:
        remarks_joined = "\n\n".join(remarks)
        blocks.append(
            Stripped(
                f"""\
<remarks>
{remarks_joined}
</remarks>"""
            )
        )

    commented_lines = [
        _slash_slash_slash_line(line) for block in blocks for line in block.splitlines()
    ]

    return Stripped("\n".join(commented_lines)), None


def generate_meta_model_comment(
    description: intermediate.MetaModelDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given meta-model."""
    return _generate_summary_remarks_constraints(description)


def generate_symbol_comment(
    description: intermediate.SymbolDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given symbol."""
    return _generate_summary_remarks_constraints(description)


def generate_property_comment(
    description: intermediate.PropertyDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    return _generate_summary_remarks_constraints(description)


def generate_enumeration_literal_comment(
    description: intermediate.EnumerationLiteralDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    return _generate_summary_remarks(description)


def generate_signature_comment(
    description: intermediate.SignatureDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the documentation comment for the given signature.

    A signature, in this context, means a function or a method signature.
    """
    errors = []  # type: List[Error]
    renderer = _ElementRenderer()

    summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )

    remarks = []  # type: List[str]
    for remark in description.remarks:
        remark, remark_errors = renderer.transform(remark)
        if remark_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark is not None
            remarks.append(remark)

    params = []  # type: List[Stripped]
    for name, body_element in description.arguments_by_name.items():
        body, body_errors = renderer.transform(body_element)
        if body_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in body_errors
            )
        else:
            assert body is not None

            # Don't use textwrap.dedent to preserve the formatting
            params.append(
                Stripped(
                    f"""\
<param name={xml.sax.saxutils.quoteattr(name)}>
{body}
</param>"""
                )
            )

    returns = None  # type: Optional[str]
    if description.returns is not None:
        # We need to help the type checker in PyCharm a bit.
        assert isinstance(description.returns, docutils.nodes.field_body)

        returns, returns_errors = renderer.transform(description.returns)
        if returns_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in returns_errors
            )
        else:
            assert returns is not None

    if len(errors) > 0:
        return None, errors

    assert summary is not None

    # Don't use textwrap.dedent to preserve the formatting

    blocks = [
        Stripped(
            f"""\
<summary>
{summary}
</summary>"""
        )
    ]

    if len(remarks) > 0:
        remarks_joined = "\n\n".join(remarks)
        blocks.append(
            Stripped(
                f"""\
<remarks>
{remarks_joined}
</remarks>"""
            )
        )

    if len(params) > 0:
        params_joined = "\n".join(params)
        blocks.append(Stripped(params_joined))

    if returns is not None:
        blocks.append(
            Stripped(
                f"""\
<returns>
{returns}
</returns>"""
            )
        )

    commented_lines = [
        _slash_slash_slash_line(line) for block in blocks for line in block.splitlines()
    ]

    return Stripped("\n".join(commented_lines)), None



"""
Infer the types of the tree nodes.

Note that these types roughly follow the type annotations in
:py:mod:`aas_core_codegen.intermediate._types`, but are not identical. For example,
the ``LENGTH`` primitive exists only in type inference. Another example, we do not
track ``parsed`` as the types are inferred in the intermediate stage, but can not
be traced back to the parse stage.
"""
import abc
import enum
from typing import Mapping, MutableMapping, Optional, List, Final, Union, get_args

from icontract import DBC, ensure

from aas_core_codegen.common import (
    Identifier,
    Error,
    assert_never,
    assert_union_of_descendants_exhaustive,
)
from aas_core_codegen.intermediate import _types
from aas_core_codegen.intermediate._types import Enumeration
from aas_core_codegen.parse import tree as parse_tree


class PrimitiveType(enum.Enum):
    """List primitive types."""

    BOOL = "bool"
    INT = "int"
    FLOAT = "float"
    STR = "str"
    BYTEARRAY = "bytearray"

    #: Denote the language-agnostic type returned from ``len(.)``.
    #: Depending on the language, this is not the same as ``INT``.
    LENGTH = "length"

    #: Denote that the node is a statement and that there is no type
    NONE = "None"


class TypeAnnotation(DBC):
    """Represent an inferred type annotation."""

    @abc.abstractmethod
    def __str__(self) -> str:
        # Signal that this is a purely abstract class
        raise NotImplementedError()


class AtomicTypeAnnotation(TypeAnnotation):
    """
    Represent an atomic type annotation.

    Atomic, in this context, means a non-generic type annotation.

    For example, ``int``.
    """

    @abc.abstractmethod
    def __str__(self) -> str:
        # Signal that this is a purely abstract class
        raise NotImplementedError()


class PrimitiveTypeAnnotation(AtomicTypeAnnotation):
    """Represent a primitive type such as ``int``."""

    def __init__(self, a_type: PrimitiveType) -> None:
        """Initialize with the given values."""
        self.a_type = a_type

    def __str__(self) -> str:
        return str(self.a_type.value)


class OurTypeAnnotation(AtomicTypeAnnotation):
    """
    Represent an atomic annotation defined by a symbol in the meta-model.

     For example, ``Asset``.
    """

    def __init__(self, symbol: _types.Symbol) -> None:
        """Initialize with the given values."""
        self.symbol = symbol

    def __str__(self) -> str:
        return self.symbol.name


class FunctionTypeAnnotation(AtomicTypeAnnotation):
    """Represent a function as a type."""

    @abc.abstractmethod
    def __str__(self) -> str:
        # Signal that this is a purely abstract class
        raise NotImplementedError()


class VerificationTypeAnnotation(FunctionTypeAnnotation):
    """Represent a type of a verification function."""

    def __init__(self, func: _types.Verification):
        """Initialize with the given values."""
        self.func = func

    def __str__(self) -> str:
        return self.func.name


class BuiltinFunction:
    """Represent a built-in function."""

    def __init__(self, name: Identifier, returns: Optional["TypeAnnotationUnion"]):
        """Initialize with the given values."""
        self.name = name
        self.returns = returns


class BuiltinFunctionTypeAnnotation(FunctionTypeAnnotation):
    """Represent a type of a built-in function."""

    def __init__(self, func: BuiltinFunction):
        """Initialize with the given values."""
        self.func = func

    def __str__(self) -> str:
        return self.func.name


class MethodTypeAnnotation(AtomicTypeAnnotation):
    """Represent a type of a class method."""

    def __init__(self, method: _types.Method):
        """Initialize with the given values."""
        self.method = method

    def __str__(self) -> str:
        return self.method.name


class SubscriptedTypeAnnotation(TypeAnnotation):
    """Represent a subscripted (i.e. generic) type annotation.

    The subscripted type annotations are, for example, ``List[...]`` (or
    ``Mapping[..., ...]``, *etc.*).
    """

    @abc.abstractmethod
    def __str__(self) -> str:
        # Signal that this is a purely abstract class
        raise NotImplementedError()


class ListTypeAnnotation(SubscriptedTypeAnnotation):
    """Represent a type annotation involving a ``List[...]``."""

    def __init__(self, items: "TypeAnnotationUnion"):
        self.items = items

    def __str__(self) -> str:
        return f"List[{self.items}]"


class OptionalTypeAnnotation(SubscriptedTypeAnnotation):
    """Represent a type annotation involving an ``Optional[...]``."""

    def __init__(self, value: "TypeAnnotationUnion"):
        self.value = value

    def __str__(self) -> str:
        return f"Optional[{self.value}]"


class EnumerationAsTypeTypeAnnotation(TypeAnnotation):
    """
    Represent an enum class as a type.

    Note that this is not the enum as a type of that enum class, but
    rather the type-as-a-type. We write``Type[T]`` in Python to describe this.
    """

    # NOTE (mristin, 2022-02-04):
    # The name of this class is admittedly clumsy. Please feel free to change if you
    # come up with a better idea.

    def __init__(self, enumeration: Enumeration) -> None:
        """Initialize with the given values."""
        self.enumeration = enumeration

    def __str__(self) -> str:
        return f"{self.__class__.__name__}[{self.enumeration}]"


def _type_annotations_equal(
    that: "TypeAnnotationUnion", other: "TypeAnnotationUnion"
) -> bool:
    """Check whether the ``that`` and ``other`` type annotations are identical."""
    if isinstance(that, PrimitiveTypeAnnotation):
        if not isinstance(other, PrimitiveTypeAnnotation):
            return False
        else:
            return that.a_type == other.a_type

    elif isinstance(that, OurTypeAnnotation):
        if not isinstance(other, OurTypeAnnotation):
            return False
        else:
            return that.symbol is other.symbol

    elif isinstance(that, VerificationTypeAnnotation):
        if not isinstance(other, VerificationTypeAnnotation):
            return False
        else:
            return that.func is other.func

    elif isinstance(that, BuiltinFunctionTypeAnnotation):
        if not isinstance(other, BuiltinFunctionTypeAnnotation):
            return False
        else:
            return that.func is other.func

    elif isinstance(that, MethodTypeAnnotation):
        if not isinstance(other, MethodTypeAnnotation):
            return False
        else:
            return that.method is other.method

    elif isinstance(that, ListTypeAnnotation):
        if not isinstance(other, ListTypeAnnotation):
            return False
        else:
            return _type_annotations_equal(that.items, other.items)

    elif isinstance(that, OptionalTypeAnnotation):
        if not isinstance(other, OptionalTypeAnnotation):
            return False
        else:
            return _type_annotations_equal(that.value, other.value)

    elif isinstance(that, EnumerationAsTypeTypeAnnotation):
        if not isinstance(other, EnumerationAsTypeTypeAnnotation):
            return False
        else:
            return that.enumeration is other.enumeration

    else:
        assert_never(that)

    raise AssertionError("Should not have gotten here")


def _assignable(
    target_type: "TypeAnnotationUnion", value_type: "TypeAnnotationUnion"
) -> bool:
    """Check whether the value can be assigned to the target."""
    if isinstance(target_type, PrimitiveTypeAnnotation):
        if isinstance(value_type, PrimitiveTypeAnnotation):
            return target_type.a_type == value_type.a_type

        # NOTE (mristin, 2021-12-25):
        # We have to be careful about the constrained primitives,
        # since we can always assign a constrained primitive to a primitive, if they
        # primitive types match.
        elif isinstance(value_type, OurTypeAnnotation) and isinstance(
            value_type.symbol, _types.ConstrainedPrimitive
        ):
            return (
                target_type.a_type == PRIMITIVE_TYPE_MAP[value_type.symbol.constrainee]
            )

        else:
            return False

    elif isinstance(target_type, OurTypeAnnotation):
        if isinstance(target_type.symbol, _types.Enumeration):
            # NOTE (mristin, 2021-12-25):
            # The enumerations are invariant.
            return (
                isinstance(value_type, OurTypeAnnotation)
                and isinstance(value_type.symbol, _types.Enumeration)
                and target_type.symbol is value_type.symbol
            )

        elif isinstance(target_type.symbol, _types.ConstrainedPrimitive):
            # NOTE (mristin, 2021-12-25):
            # If it is a constrained primitive with no constraints, allow the assignment
            # if the target and the value match on the primitive type.
            if len(target_type.symbol.invariants) == 0 and isinstance(
                value_type, PrimitiveTypeAnnotation
            ):
                return (
                    PRIMITIVE_TYPE_MAP[target_type.symbol.constrainee]
                    == value_type.a_type
                )
            else:
                # NOTE (mristin, 2021-12-25):
                # We assume the assignments of constrained primitives to be co-variant.
                if (
                    isinstance(value_type, OurTypeAnnotation)
                    and isinstance(value_type.symbol, _types.ConstrainedPrimitive)
                    and target_type.symbol.constrainee == value_type.symbol.constrainee
                ):
                    return (
                        target_type.symbol is value_type.symbol
                        or id(value_type.symbol) in target_type.symbol.descendant_id_set
                    )

            return False

        elif isinstance(target_type.symbol, _types.Class):
            if not (
                isinstance(value_type, OurTypeAnnotation)
                and isinstance(value_type.symbol, _types.Class)
            ):
                return False

            # NOTE (mristin, 2021-12-25):
            # We assume the assignment to be co-variant. Either the target type and
            # the value type are equal *or* the value symbol is a descendant of the
            # target symbol.

            return target_type.symbol is value_type.symbol or (
                id(value_type.symbol) in target_type.symbol.descendant_id_set
            )

    elif isinstance(target_type, VerificationTypeAnnotation):
        if not isinstance(value_type, VerificationTypeAnnotation):
            return False
        else:
            return target_type.func is value_type.func

    elif isinstance(target_type, BuiltinFunctionTypeAnnotation):
        if not isinstance(value_type, BuiltinFunctionTypeAnnotation):
            return False
        else:
            return target_type.func is value_type.func

    elif isinstance(target_type, MethodTypeAnnotation):
        if not isinstance(value_type, MethodTypeAnnotation):
            return False
        else:
            return target_type.method is value_type.method

    elif isinstance(target_type, ListTypeAnnotation):
        if not isinstance(value_type, ListTypeAnnotation):
            return False
        else:
            # NOTE (mristin, 2021-12-25):
            # We assume the lists to be invariant. This is necessary for code generation
            # in implementation targets such as C++ and Golang.
            return _type_annotations_equal(target_type.items, value_type.items)

    elif isinstance(target_type, OptionalTypeAnnotation):
        # NOTE (mristin, 2021-12-25):
        # We can always assign a non-optional to an optional.
        if not isinstance(value_type, OptionalTypeAnnotation):
            return _assignable(target_type=target_type.value, value_type=value_type)
        else:
            # NOTE (mristin, 2021-12-25):
            # We assume the optionals to be co-variant.
            return _assignable(
                target_type=target_type.value, value_type=value_type.value
            )

    elif isinstance(target_type, EnumerationAsTypeTypeAnnotation):
        raise NotImplementedError(
            "(mristin, 2022-02-04): Assigning enumeration-as-type to another "
            "enumeration-as-type is a very niche program logic. As we do not have "
            "a concrete example of such an assignment, we currently ignore this case "
            "in determining whether the assignment makes sense. When you have "
            "a concrete example, please revisit this part of the code."
        )

    else:
        assert_never(target_type)

    return False


PRIMITIVE_TYPE_MAP = {
    _types.PrimitiveType.BOOL: PrimitiveType.BOOL,
    _types.PrimitiveType.INT: PrimitiveType.INT,
    _types.PrimitiveType.FLOAT: PrimitiveType.FLOAT,
    _types.PrimitiveType.STR: PrimitiveType.STR,
    _types.PrimitiveType.BYTEARRAY: PrimitiveType.BYTEARRAY,
}

for _types_primitive_type in _types.PrimitiveType:
    assert (
        _types_primitive_type in PRIMITIVE_TYPE_MAP
    ), f"All primitive types from _types covered, but: {_types_primitive_type=}"


def _type_annotation_to_inferred_type_annotation(
    type_annotation: _types.TypeAnnotationUnion,
) -> "TypeAnnotationUnion":
    """Convert from the :py:mod:`aas_core_codegen.intermediate._types`."""
    if isinstance(type_annotation, _types.PrimitiveTypeAnnotation):
        return PrimitiveTypeAnnotation(
            a_type=PRIMITIVE_TYPE_MAP[type_annotation.a_type]
        )

    elif isinstance(type_annotation, _types.OurTypeAnnotation):
        return OurTypeAnnotation(symbol=type_annotation.symbol)

    elif isinstance(type_annotation, _types.ListTypeAnnotation):
        return ListTypeAnnotation(
            items=_type_annotation_to_inferred_type_annotation(type_annotation.items)
        )

    elif isinstance(type_annotation, _types.OptionalTypeAnnotation):
        return OptionalTypeAnnotation(
            value=_type_annotation_to_inferred_type_annotation(type_annotation.value)
        )

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


class Environment(DBC):
    """
    Map names to type annotations for a given scope.

    We first search in the given scope and then iterate over the ancestor scopes.
    See, for example: https://craftinginterpreters.com/resolving-and-binding.html.

    The most outer, global, scope is parentless.
    """

    @property
    @abc.abstractmethod
    def mapping(self) -> Mapping[Identifier, "TypeAnnotationUnion"]:
        """Retrieve the underlying mapping."""
        raise NotImplementedError()

    def __init__(self, parent: Optional["Environment"]) -> None:
        """Initialize with the given values."""
        self.parent = parent

    def find(self, identifier: Identifier) -> Optional["TypeAnnotationUnion"]:
        """
        Search for the type annotation of the given ``identifier``.

        We search all the way to the most outer scope.
        """
        type_anno = self.mapping.get(identifier, None)
        if type_anno is not None:
            return type_anno

        if self.parent is not None:
            return self.parent.find(identifier)

        return None


class ImmutableEnvironment(Environment):
    """
    Map immutably names to type annotations for a given scope.
    """

    @property
    def mapping(self) -> Mapping[Identifier, "TypeAnnotationUnion"]:
        """Retrieve the underlying mapping."""
        return self._mapping

    def __init__(
        self,
        mapping: Mapping[Identifier, "TypeAnnotationUnion"],
        parent: Optional["Environment"] = None,
    ) -> None:
        self._mapping = mapping

        Environment.__init__(self, parent)


class MutableEnvironment(Environment):
    """
    Map names to type annotations for a given scope and allow mutations.
    """

    @property
    def mapping(self) -> Mapping[Identifier, "TypeAnnotationUnion"]:
        """Retrieve the underlying mapping."""
        return self._mapping

    def __init__(self, parent: Optional["Environment"] = None) -> None:
        self._mapping = (
            dict()
        )  # type: MutableMapping[Identifier, "TypeAnnotationUnion"]

        Environment.__init__(self, parent)

    def set(
        self, identifier: Identifier, type_annotation: "TypeAnnotationUnion"
    ) -> None:
        """Set the ``type_annotation`` for the given ``identifier``."""
        self._mapping[identifier] = type_annotation


class Inferrer(parse_tree.RestrictedTransformer[Optional["TypeAnnotationUnion"]]):
    """Infer the types of the given parse tree."""

    #: Track of the inferred types
    type_map: Final[MutableMapping[parse_tree.Node, "TypeAnnotationUnion"]]

    #: Errors encountered during the inference
    errors: Final[List[Error]]

    def __init__(
        self,
        symbol_table: _types.SymbolTable,
        environment: "Environment",
    ) -> None:
        """Initialize with the given values."""
        # We need to create our own child environment so that we can introduce new
        # entries without affecting the variables from the outer scopes.
        self._environment = MutableEnvironment(parent=environment)

        self._symbol_table = symbol_table

        self.type_map = dict()
        self.errors = []

    @ensure(lambda self, result: not (result is None) or len(self.errors) > 0)
    def transform(self, node: parse_tree.Node) -> Optional["TypeAnnotationUnion"]:
        return super().transform(node)

    def transform_member(
        self, node: parse_tree.Member
    ) -> Optional["TypeAnnotationUnion"]:
        instance_type = self.transform(node.instance)

        if instance_type is None:
            return None

        if isinstance(instance_type, OurTypeAnnotation) and isinstance(
            instance_type.symbol, _types.Class
        ):
            cls = instance_type.symbol
            assert isinstance(cls, _types.Class)

            prop = cls.properties_by_name.get(node.name, None)
            if prop is not None:
                result = _type_annotation_to_inferred_type_annotation(
                    prop.type_annotation
                )
                self.type_map[node] = result
                return result

            method = cls.methods_by_name.get(node.name, None)
            if method is not None:
                result = MethodTypeAnnotation(method=method)
                self.type_map[node] = result
                return result

            self.errors.append(
                Error(
                    node.original_node,
                    f"The member {node.name!r} could not be found "
                    f"in the class {cls.name!r}",
                )
            )

        elif isinstance(instance_type, EnumerationAsTypeTypeAnnotation):
            enumeration = instance_type.enumeration
            literal = enumeration.literals_by_name.get(node.name, None)
            if literal is not None:
                result = OurTypeAnnotation(symbol=enumeration)
                self.type_map[node] = result
                return result

            self.errors.append(
                Error(
                    node.original_node,
                    f"The literal {node.name!r} could not be found "
                    f"in the enumeration {enumeration.name!r}",
                )
            )
        else:
            self.errors.append(
                Error(
                    node.original_node,
                    f"Expected an instance type to be either an enumeration-as-type or "
                    f"a class, but got: {instance_type}",
                )
            )
            return None

        return None

    def transform_comparison(
        self, node: parse_tree.Comparison
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``left`` and ``right`` even though we
        # know the type in advance
        success = (self.transform(node.left) is not None) and (
            self.transform(node.right) is not None
        )

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_implication(
        self, node: parse_tree.Implication
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``antecedent`` and ``consequent`` even
        # though we know the type in advance
        success = (self.transform(node.antecedent) is not None) and (
            self.transform(node.consequent) is not None
        )

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_method_call(
        self, node: parse_tree.MethodCall
    ) -> Optional["TypeAnnotationUnion"]:
        # Simply recurse to track the type, but we don't care about the arguments
        failed = False
        for arg in node.args:
            arg_type = self.transform(arg)
            if arg_type is None:
                failed = True

        member_type = self.transform(node.member)

        if member_type is None:
            failed = True

        if failed:
            return None

        if not isinstance(member_type, MethodTypeAnnotation):
            self.errors.append(
                Error(
                    node.original_node,
                    f"Expected the member in a method call to be a method, "
                    f"but got: {member_type}",
                )
            )
            return None

        # noinspection PyUnusedLocal
        result = None  # type: Optional[TypeAnnotationUnion]

        if member_type.method.returns is None:
            result = PrimitiveTypeAnnotation(a_type=PrimitiveType.NONE)
        else:
            result = _type_annotation_to_inferred_type_annotation(
                member_type.method.returns
            )

        assert result is not None
        self.type_map[node] = result
        return result

    def transform_function_call(
        self, node: parse_tree.FunctionCall
    ) -> Optional["TypeAnnotationUnion"]:
        result = None  # type: Optional[TypeAnnotationUnion]
        failed = False

        func_type = self.transform(node.name)
        if func_type is None:
            failed = True
        else:
            # NOTE (mristin, 2021-12-26):
            # The verification functions use
            # :py:mod:`aas_core_codegen.intermediate._types` while the built-in
            # functions are a construct of
            # :py:mod:`aas_core_codegen.intermediate.type_inference`.

            if isinstance(func_type, VerificationTypeAnnotation):
                if func_type.func.returns is not None:
                    result = _type_annotation_to_inferred_type_annotation(
                        func_type.func.returns
                    )
                else:
                    result = PrimitiveTypeAnnotation(PrimitiveType.NONE)

            elif isinstance(func_type, BuiltinFunctionTypeAnnotation):
                if func_type.func.returns is not None:
                    result = func_type.func.returns
                else:
                    result = PrimitiveTypeAnnotation(PrimitiveType.NONE)

            elif isinstance(
                func_type,
                (
                    PrimitiveTypeAnnotation,
                    OurTypeAnnotation,
                    MethodTypeAnnotation,
                    ListTypeAnnotation,
                    OptionalTypeAnnotation,
                    EnumerationAsTypeTypeAnnotation,
                ),
            ):
                self.errors.append(
                    Error(
                        node.name.original_node,
                        f"Expected the variable {node.name.identifier!r} to be "
                        f"a function, but got {func_type}",
                    )
                )

            else:
                assert_never(func_type)

        # NOTE (mristin, 2021-12-26):
        # Recurse to track the type of arguments. Even if we failed before, we want to
        # catch the errors in the arguments for better developer experience.
        #
        # Mind that we are sloppy here. Theoretically, we could check that arguments in
        # the call are assignable to the arguments in the function definition and catch
        # errors in the meta-model at this point. However, we prioritize other features
        # and leave this check unimplemented for now.

        for arg in node.args:
            arg_type = self.transform(arg)
            if arg_type is None:
                failed = True

        if failed:
            return None

        assert result is not None
        self.type_map[node] = result
        return result

    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Optional["TypeAnnotationUnion"]:
        result = None  # type: Optional[TypeAnnotationUnion]

        if isinstance(node.value, bool):
            result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        elif isinstance(node.value, int):
            result = PrimitiveTypeAnnotation(PrimitiveType.INT)
        elif isinstance(node.value, float):
            result = PrimitiveTypeAnnotation(PrimitiveType.FLOAT)
        elif isinstance(node.value, str):
            result = PrimitiveTypeAnnotation(PrimitiveType.STR)
        else:
            assert_never(node.value)

        assert result is not None

        self.type_map[node] = result
        return result

    def transform_is_none(
        self, node: parse_tree.IsNone
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``value`` even though we know the type in
        # advance
        success = self.transform(node.value) is not None

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_is_not_none(
        self, node: parse_tree.IsNotNone
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``value`` even though we know the type in
        # advance
        success = self.transform(node.value) is not None

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_name(self, node: parse_tree.Name) -> Optional["TypeAnnotationUnion"]:
        type_in_env = self._environment.find(node.identifier)
        if type_in_env is None:
            self.errors.append(
                Error(
                    node.original_node,
                    f"We do not know how to infer the type of "
                    f"the variable with the identifier {node.identifier!r} from the "
                    f"given environment. Mind that we do not consider the module "
                    f"scope nor handle all built-in functions due to simplicity! If "
                    f"you believe this needs to work, please notify the developers.",
                )
            )
            return None

        self.type_map[node] = type_in_env
        return type_in_env

    def transform_and(self, node: parse_tree.And) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``values`` even though we know the type
        # in advance
        success = all(
            self.transform(value_node) is not None for value_node in node.values
        )

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_or(self, node: parse_tree.Or) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``values`` even though we know the type
        # in advance
        success = all(
            self.transform(value_node) is not None for value_node in node.values
        )

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_formatted_value(
        self, node: parse_tree.FormattedValue
    ) -> Optional["TypeAnnotationUnion"]:
        value_type = self.transform(node.value)
        if value_type is None:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.STR)
        self.type_map[node] = result
        return result

    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``values`` even though we know the type
        # in advance
        success = True
        for value in node.values:
            if isinstance(value, str):
                continue
            elif isinstance(value, parse_tree.FormattedValue):
                formatted_value_type = self.transform(value)
                if formatted_value_type is None:
                    success = False
            else:
                assert_never(value)

        if not success:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.STR)
        self.type_map[node] = result
        return result

    def transform_for_each(
        self, node: parse_tree.ForEach
    ) -> Optional["TypeAnnotationUnion"]:
        variable_type_in_env = self._environment.find(node.variable.identifier)
        if variable_type_in_env is not None:
            self.errors.append(
                Error(
                    node.variable.original_node,
                    f"The variable {node.variable.identifier} "
                    f"has been already defined before",
                )
            )
            return None

        iter_type = self.transform(node.iteration)
        if iter_type is None:
            return None

        while isinstance(iter_type, OptionalTypeAnnotation):
            iter_type = iter_type.value

        if not isinstance(iter_type, ListTypeAnnotation):
            self.errors.append(
                Error(
                    node.iteration.original_node,
                    f"Expected an iteration over a list, but got: {iter_type}",
                )
            )
            return None

        self._environment.set(
            identifier=node.variable.identifier, type_annotation=iter_type.items
        )

        result = PrimitiveTypeAnnotation(PrimitiveType.NONE)
        self.type_map[node] = result
        return result

    def transform_all(self, node: parse_tree.All) -> Optional["TypeAnnotationUnion"]:
        a_type = self.transform(node.for_each)
        if a_type is None:
            return None

        a_type = self.transform(node.condition)
        if a_type is None:
            return None

        result = PrimitiveTypeAnnotation(PrimitiveType.BOOL)
        self.type_map[node] = result
        return result

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Optional["TypeAnnotationUnion"]:
        is_new_variable = False

        # noinspection PyUnusedLocal
        target_type = None  # type: Optional[TypeAnnotationUnion]

        if isinstance(node.target, parse_tree.Name):
            target_type = self._environment.find(node.target.identifier)
            if target_type is None:
                is_new_variable = True
        else:
            target_type = self.transform(node.target)

        value_type = self.transform(node.value)

        if (not is_new_variable and target_type is None) or (value_type is None):
            return None

        if target_type is not None and not _assignable(
            target_type=target_type, value_type=value_type
        ):
            self.errors.append(
                Error(
                    node.original_node,
                    f"We inferred the target type of the assignment to "
                    f"be {target_type}, while the value type is inferred to "
                    f"be {value_type}. We do not know how to model this assignment.",
                )
            )
            return None

        if is_new_variable:
            assert isinstance(node.target, parse_tree.Name)

            self._environment.set(
                identifier=node.target.identifier, type_annotation=value_type
            )

        result = PrimitiveTypeAnnotation(PrimitiveType.NONE)
        self.type_map[node] = result
        return result

    def transform_return(
        self, node: parse_tree.Return
    ) -> Optional["TypeAnnotationUnion"]:
        # Just recurse to fill ``type_map`` on ``value`` even though we know the type
        # in advance
        if node.value is not None:
            success = self.transform(node.value) is not None
            if not success:
                return None

        # Treat ``return`` as a statement
        result = PrimitiveTypeAnnotation(PrimitiveType.NONE)
        self.type_map[node] = result
        return result


def populate_base_environment(symbol_table: _types.SymbolTable) -> Environment:
    """Create a basic mapping name 🠒 type annotation from the global scope.

    The global scope, in this context, refers to the level of symbol table.
    """
    # Build up the environment;
    # see https://craftinginterpreters.com/resolving-and-binding.html
    mapping: MutableMapping[Identifier, "TypeAnnotationUnion"] = {
        Identifier("len"): BuiltinFunctionTypeAnnotation(
            func=BuiltinFunction(
                name=Identifier("len"),
                returns=PrimitiveTypeAnnotation(PrimitiveType.LENGTH),
            )
        )
    }

    for verification in symbol_table.verification_functions:
        assert verification.name not in mapping
        mapping[verification.name] = VerificationTypeAnnotation(func=verification)

    for symbol in symbol_table.symbols:
        if isinstance(symbol, _types.Enumeration):
            assert symbol.name not in mapping
            mapping[symbol.name] = EnumerationAsTypeTypeAnnotation(enumeration=symbol)

    return ImmutableEnvironment(mapping=mapping, parent=None)


TypeAnnotationUnion = Union[
    PrimitiveTypeAnnotation,
    OurTypeAnnotation,
    VerificationTypeAnnotation,
    BuiltinFunctionTypeAnnotation,
    MethodTypeAnnotation,
    ListTypeAnnotation,
    OptionalTypeAnnotation,
    EnumerationAsTypeTypeAnnotation,
]
assert_union_of_descendants_exhaustive(
    union=TypeAnnotationUnion, base_class=TypeAnnotation
)

FunctionTypeAnnotationUnion = Union[
    VerificationTypeAnnotation, BuiltinFunctionTypeAnnotation
]
assert_union_of_descendants_exhaustive(
    union=FunctionTypeAnnotationUnion, base_class=FunctionTypeAnnotation
)

# NOTE (mristin, 2021-12-27):
# Mypy is not smart enough to work with ``get_args``, so we have to manually write it
# out.
FunctionTypeAnnotationUnionAsTuple = (
    VerificationTypeAnnotation,
    BuiltinFunctionTypeAnnotation,
)
assert FunctionTypeAnnotationUnionAsTuple == get_args(FunctionTypeAnnotationUnion)



"""Provide common functions and types for the code generation."""
import ast
import inspect
import io
import re
import textwrap
from typing import (
    Optional,
    Tuple,
    cast,
    List,
    overload,
    Union,
    Iterator,
    Sequence,
    NoReturn,
    Any,
    Iterable,
)

import asttokens
from icontract import require, DBC

IDENTIFIER_RE = re.compile(r"[a-zA-Z_][a-zA-Z_0-9]*")


class Identifier(DBC, str):
    """Represent an identifier."""

    @require(lambda value: IDENTIFIER_RE.fullmatch(value))
    def __new__(cls, value: str) -> "Identifier":
        return cast(Identifier, value)


class Error:
    """
    Represent an unexpected input.

    For example, the code of the meta-model can be a valid Python code, but we
    can only process a subset of language constructs.
    """

    def __init__(
        self,
        node: Optional[ast.AST],
        message: str,
        underlying: Optional[List["Error"]] = None,
    ) -> None:
        self.node = node
        self.message = message
        self.underlying = underlying

    def __repr__(self) -> str:
        return (
            f"Error("
            f"node={self.node!r}, "
            f"message={self.message!r}, "
            f"underlying={self.underlying!r})"
        )


class LinenoColumner:
    """Map the source code to line number and column for precise error messages."""

    def __init__(self, atok: asttokens.ASTTokens) -> None:
        positions = []  # type: List[Tuple[int, int]]
        lineno = 1
        column = 0
        for character in atok.get_text(atok.tree):
            if character == "\n":
                column = 1
                lineno += 1
            else:
                column += 1

            positions.append((lineno, column))

        self.atok = atok
        self.positions = positions

    def error_message(self, error: Error) -> str:
        """Generate the error message based on the unexpected observation."""
        prefix = ""
        if error.node is not None:
            start, _ = self.atok.get_text_range(node=error.node)
            lineno, column = self.positions[start]

            prefix = f"At line {lineno} and column {column}: "

        if error.underlying is None or len(error.underlying) == 0:
            return f"{prefix}{error.message}"
        else:
            writer = io.StringIO()
            writer.write(f"{prefix}{error.message}\n")
            for i, underlying_error in enumerate(error.underlying):
                if i > 0:
                    writer.write("\n")
                indented = textwrap.indent(self.error_message(underlying_error), "  ")
                writer.write(indented)

            return writer.getvalue()


class Lines(DBC):
    """Represent a sequence of text lines."""

    # fmt: off
    @require(
        lambda lines:
        all('\n' not in line and '\r' not in line for line in lines)
    )
    # fmt: on
    def __new__(cls, lines: Sequence[str]) -> "Lines":
        r"""
        Ensure the properties on the ``lines``.

        Please make sure that you transfer the "ownership" immediately to Lines
        and don't modify the original list of strings any more:

        .. code-block:: python

            ##
            # OK
            ##

            lines = Lines(some_text.splitlines())

            ##
            # Not OK
            ##

            some_lines = some_text.splitlines()
            lines = Lines(some_lines)
            # ... do something assuming ``lines`` is immutable ...

            some_lines[0] = "This will break \n your logic"
            # ERROR! lines[0] now contains a new-line which is unexpected!

        """
        return cast(Lines, lines)

    def __add__(self, other: "Lines") -> "Lines":
        """Concatenate two list of lines."""
        raise NotImplementedError("Only for type annotations")

    # pylint: disable=function-redefined

    @overload
    def __getitem__(self, index: int) -> str:
        """Get the item at the given integer index."""
        raise NotImplementedError("Only for type annotations")

    @overload
    def __getitem__(self, index: slice) -> "Lines":
        """Get the slice of the lines."""
        raise NotImplementedError("Only for type annotations")

    def __getitem__(self, index: Union[int, slice]) -> Union[str, "Lines"]:
        """Get the line(s) at the given index."""
        raise NotImplementedError("Only for type annotations")

    def __len__(self) -> int:
        """Return the number of the lines."""
        raise NotImplementedError("Only for type annotations")

    def __iter__(self) -> Iterator[str]:
        """Iterate over the lines."""
        raise NotImplementedError("Only for type annotations")


def assert_never(value: NoReturn) -> NoReturn:
    """
    Signal to mypy to perform an exhaustive matching.

    Please see the following page for more details:
    https://hakibenita.com/python-mypy-exhaustive-checking
    """
    assert False, f"Unhandled value: {value} ({type(value).__name__})"


class Rstripped(str):
    """
    Represent a block of text without trailing whitespace.

    The block can be both single-line or multi-line.
    """

    @require(
        lambda block: not block.endswith("\n")
        and not block.endswith(" ")
        and not block.endswith("\t")
    )
    def __new__(cls, block: str) -> "Rstripped":
        return cast(Rstripped, block)


class Stripped(Rstripped):
    """
    Represent a block of text without leading and trailing whitespace.

    The block of text can be both single-line and multi-line.
    """

    # fmt: off
    @require(
        lambda block:
        not block.startswith('\n')
        and not block.startswith(' ')
        and not block.startswith('\t')
    )
    @require(
        lambda block:
        not block.endswith('\n')
        and not block.endswith(' ')
        and not block.endswith('\t')
    )
    # fmt: on
    def __new__(cls, block: str) -> "Stripped":
        return cast(Stripped, block)


def indent_but_first_line(text: str, indention: str) -> str:
    """
    Indent all but the first of the given ``text`` by ``indention``.

    For example, this helps you insert indented blocks into formatted string literals.
    """
    indented_lines = []  # type: List[str]
    for i, line in enumerate(text.splitlines()):
        if i == 0:
            indented_lines.append(line)
        else:
            if len(line) > 0:
                indented_lines.append(indention + line)
            else:
                indented_lines.append(line)

    return "\n".join(indented_lines)


def assert_union_of_descendants_exhaustive(union: Any, base_class: Any) -> None:
    """
    Check that the ``union`` covers all the concrete subclasses of ``base_class``.

    Make sure you put the assertion at the end of the module where no new classes are
    defined.

    See also for more details: https://hakibenita.com/python-mypy-exhaustive-checking
    """
    if inspect.isclass(union):
        union_map = {id(union): union}
    elif hasattr(union, "__args__"):
        union_map = {id(cls): cls for cls in union.__args__}
    else:
        raise NotImplementedError(f"We do not know how to handle the union: {union}")

    # We have to recursively figure out the sub-classes.
    concrete_subclasses = []  # type: List[Any]

    stack = base_class.__subclasses__()  # type: List[Any]

    while len(stack) > 0:
        sub_cls = stack.pop()
        if not inspect.isabstract(sub_cls):
            concrete_subclasses.append(sub_cls)

        stack.extend(sub_cls.__subclasses__())

    subclass_map = {id(sub_cls): sub_cls for sub_cls in concrete_subclasses}

    union_set = set(union_map.keys())
    subclass_set = set(subclass_map.keys())

    if union_set != subclass_set:
        union_diff = union_set.difference(subclass_set)
        union_diff_names = [union_map[cls_id].__name__ for cls_id in union_diff]

        subclass_diff = subclass_set.difference(union_set)
        subclass_diff_names = [
            subclass_map[cls_id].__name__ for cls_id in subclass_diff
        ]

        if len(union_diff_names) == 0 and len(subclass_diff_names) > 0:
            raise AssertionError(
                f"The following concrete subclasses of {base_class.__name__!r} were "
                f"not listed in the union: {subclass_diff_names}"
            )

        elif len(union_diff_names) > 0 and len(subclass_diff_names) == 0:
            raise AssertionError(
                f"The following classes were listed in the union, "
                f"but they are not sub-classes "
                f"of {base_class.__name__!r}: {union_diff_names}"
            )
        else:
            raise AssertionError(
                f"The following classes were listed in the union, "
                f"but they are not sub-classes "
                f"of {base_class.__name__!r}: {union_diff_names}.\n\n"
                f"The following concrete sub-classes of {base_class.__name__!r} were "
                f"not listed in the union: {subclass_diff_names}"
            )


def assert_union_without_excluded(
    original_union: Any, subset_union: Any, excluded: Iterable[Any]
) -> None:
    """
    Check that the ``subset_union`` ∪ ``excluded`` is ``original_union``.

    Make sure you put the assertion at the end of the module where no new classes are
    defined.

    See also for more details: https://hakibenita.com/python-mypy-exhaustive-checking
    """
    # region Map the identifiers of the inputs to their objects

    if inspect.isclass(original_union):
        original_union_map = {id(original_union): original_union}
    elif hasattr(original_union, "__args__"):
        original_union_map = {id(a_type): a_type for a_type in original_union.__args__}
    else:
        raise NotImplementedError(
            f"We do not know how to handle the original_union: {original_union}"
        )

    if inspect.isclass(subset_union):
        subset_union_map = {id(subset_union): subset_union}
    elif hasattr(subset_union, "__args__"):
        subset_union_map = {id(a_type): a_type for a_type in subset_union.__args__}
    else:
        raise NotImplementedError(
            f"We do not know how to handle the subset_union: {subset_union}"
        )

    excluded_map = {id(a_type): a_type for a_type in excluded}

    name_map = {**original_union_map, **subset_union_map, **excluded_map}

    # endregion

    # region Compute the sets of the identifiers

    original_union_set = set(original_union_map.keys())
    subset_union_set = set(subset_union_map.keys())
    excluded_set = set(excluded_map.keys())

    # endregion

    # region Check that it all fits

    intersection = subset_union_set.intersection(excluded_set)
    if len(intersection) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in intersection)
        raise AssertionError(
            f"The following types were listed both "
            f"in the subset_union and the excluded: {names}"
        )

    diff = subset_union_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the subset_union, "
            f"but not in the original_union: {names}"
        )

    diff = excluded_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the excluded, "
            f"but not in the original_union: {names}"
        )

    reconstructed_set = subset_union_set.union(excluded_set)

    diff = original_union_set.difference(reconstructed_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the original_union, "
            f"but not in the subset_union ∪ excluded: {names}"
        )

    diff = reconstructed_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the subset_union ∪ excluded, "
            f"but not in the original_union: {names}"
        )

    # endregion



# pylint: disable=missing-docstring

import unittest

from aas_core_codegen.rdf_shacl import common as rdf_shacl_common


class Test_string_literal(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual('""', rdf_shacl_common.string_literal(""))

    def test_no_escapes(self) -> None:
        self.assertEqual('"something"', rdf_shacl_common.string_literal("something"))

    def test_escape_single_line(self) -> None:
        self.assertEqual(
            '"some\\"thing"', rdf_shacl_common.string_literal('some"thing')
        )

    def test_multi_line_no_escape(self) -> None:
        self.assertEqual(
            r'"some\nthing"', rdf_shacl_common.string_literal("some\nthing")
        )

    def test_multi_line_escape(self) -> None:
        self.assertEqual(
            r'"some\nthi\"ng"', rdf_shacl_common.string_literal('some\nthi"ng')
        )


if __name__ == "__main__":
    unittest.main()



"""Provide common functions shared among difference C# code generation modules."""
import re
import textwrap
from typing import List, Union, cast, Iterator

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never
from aas_core_codegen.csharp import naming as csharp_naming


@ensure(lambda result: result.startswith('"'))
@ensure(lambda result: result.endswith('"'))
def string_literal(text: str) -> str:
    """Generate a C# string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        if character == "\a":
            escaped.append("\\a")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\t":
            escaped.append("\\t")
        elif character == "\v":
            escaped.append("\\v")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        else:
            escaped.append(character)

    return '"{}"'.format("".join(escaped))


def needs_escaping(text: str) -> bool:
    """Check whether the ``text`` contains a character that needs escaping."""
    for character in text:
        if character == "\a":
            return True
        elif character == "\b":
            return True
        elif character == "\f":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\t":
            return True
        elif character == "\v":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        else:
            pass

    return False


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("bool"),
    intermediate.PrimitiveType.INT: Stripped("long"),
    intermediate.PrimitiveType.FLOAT: Stripped("double"),
    intermediate.PrimitiveType.STR: Stripped("string"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("byte[]"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that we have explicitly mapped all the primitive types to C#."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()


def generate_type(type_annotation: intermediate.TypeAnnotationUnion) -> Stripped:
    """Generate the C# type for the given type annotation."""
    # BEFORE-RELEASE (mristin, 2021-12-13): test in isolation
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        symbol = type_annotation.symbol

        if isinstance(symbol, intermediate.Enumeration):
            return Stripped(csharp_naming.enum_name(type_annotation.symbol.name))

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[symbol.constrainee]

        elif isinstance(symbol, intermediate.Class):
            # NOTE (mristin, 2021-12-26):
            # Always prefer an interface to allow for discrimination. If there is
            # an interface based on the class, it means that there are one or more
            # descendants.

            if symbol.interface:
                return Stripped(csharp_naming.interface_name(symbol.name))
            else:
                return Stripped(csharp_naming.class_name(symbol.name))

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(type_annotation=type_annotation.items)

        return Stripped(f"List<{item_type}>")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value = generate_type(type_annotation=type_annotation.value)
        return Stripped(f"{value}?")

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT = "    "
INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6

NAMESPACE_IDENTIFIER_RE = re.compile(
    r"[a-zA-Z_][a-zA-Z_0-9]*(\.[a-zA-Z_][a-zA-Z_0-9]*)"
)


class NamespaceIdentifier:
    """Capture a namespace identifier."""

    @require(lambda identifier: NAMESPACE_IDENTIFIER_RE.fullmatch(identifier))
    def __new__(cls, identifier: str) -> "NamespaceIdentifier":
        return cast(NamespaceIdentifier, identifier)


WARNING = Stripped(
    textwrap.dedent(
        """\
    /*
     * This code has been automatically generated by aas-core-codegen.
     * Do NOT edit or append.
     */"""
    )
)


def over_enumerations_classes_and_interfaces(
    symbol_table: intermediate.SymbolTable,
) -> Iterator[
    Union[intermediate.Enumeration, intermediate.ConcreteClass, intermediate.Interface]
]:
    """
    Iterate over all enumerations, concrete classes and interfaces.

    These intermediate structures form the base of the C# code.
    """
    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            yield symbol
        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            pass
        elif isinstance(symbol, intermediate.AbstractClass):
            yield symbol.interface
        elif isinstance(symbol, intermediate.ConcreteClass):
            if symbol.interface:
                yield symbol.interface

            yield symbol
        else:
            assert_never(symbol)
